from orwynn.di.collecting.errors import (
    NoDependenciesForGivenProviderError,
    ProviderAlreadyInitializedForMapError,
    ProviderNotFoundInMapError,
)
from orwynn.di.provider import Provider


class ProviderDependenciesMap:
    """Maps Providers and their requested dependencies.

    Since providers can request only other providers, it's a representation of
    Provider to Provider map.
    """
    def __init__(
        self
    ) -> None:
        self._map: dict[type[Provider], list[type[Provider]]] = {}

    @property
    def Providers(self) -> list[type[Provider]]:
        return list(self._map.keys())

    @property
    def mapped_items(self):
        return self._map.items()

    def get_dependencies(self, P: type[Provider]) -> list[type[Provider]]:
        """Gets dependencies for given provider.

        Args:
            P:
                Provider to fetch dependencies from.

        Returns:
            List of dependencies of provider.

        Raises:
            ProviderNotFoundInMapError:
                Provider is not found.
            NoDependenciesForProviderError:
                No dependencies for given Provider is found.
        """
        res = self._map.get(P, None)
        if res:
            return res
        elif res == []:
            raise NoDependenciesForGivenProviderError(P=P)
        else:
            raise ProviderNotFoundInMapError(P=P)

    def init_provider(
        self,
        P: type[Provider]
    ) -> None:
        """Initialize given Provider in the map.

        Args:
            P:
                Provider to be initialized.

        Raises:
            ProviderAlreadyInitializedForMapError:
                Provider already exists in the map.
        """
        if P in self._map:
            raise ProviderAlreadyInitializedForMapError(FailedProvider=P)
        self._map[P] = []

    def add_dependency(
        self, *, dependency: type[Provider], P: type[Provider]
    ) -> None:
        """Adds dependency for Provider.

        Creates a new Provider entry in the map if it's not exist.

        Args:
            dependency:
                Dependency to add.
            P:
                Provider to add to.
        """
        if P in self._map:
            self._map[P].append(dependency)
        else:
            self._map[P] = [dependency]



"""
Works with url formatting.
"""
import re
from enum import Enum
from itertools import zip_longest

from pykit import validation
from pykit.cls import Static
from starlette.datastructures import URL as _StarletteURL

from orwynn.model.model import Model

URL = _StarletteURL


class URLScheme(Enum):
    """
    Supported URI schemes.
    """
    HTTP = "http"
    HTTPS = "https"
    Websocket = "websocket"
    RTSP = "rtsp"


class URLVars(Model):
    url: URL
    abstract_route: str
    path_vars: dict[str, str]
    query_vars: dict[str, str]

    class Config:
        arbitrary_types_allowed = True


class URLUtils(Static):
    @staticmethod
    def join_routes(*routes: str) -> str:
        """Joins all given routes and normalize final result."""
        validation.validate_each(routes, str, expected_sequence_type=tuple)

        result: str = ""

        for route in routes:
            if route == "" or route == "/":
                continue
            elif route[0] != "/":
                result += "/" + route
            else:
                result += route
            result.removesuffix("/")

        if result == "":
            result = "/"

        return result

    @staticmethod
    def get_vars(
        url: URL,
        *,
        abstract_route: str
    ) -> URLVars:
        """
        Inspects the given url to find all path and query vars.

        Args:
            url:
                Url to inspect.
            abstract_route:
                Route with formatting blocks for route vars included.

        Returns:
            Dict formed of path and query vars.
        """
        final: URLVars = URLVars(
            url=url,
            abstract_route=abstract_route,
            path_vars={},
            query_vars={}
        )
        URLUtils._set_path_vars(final)
        if url.query != "":
            URLUtils._set_query_vars(final)
        return final

    @staticmethod
    def _set_path_vars(
        url_vars: URLVars
    ) -> None:
        path_var_names: list[str] = re.findall(
            r"\{\w+\}",
            url_vars.abstract_route
        )

        # Remove brackets from each element in acquired path variable names
        path_var_names = [
            name.replace("{", "").replace("}", "") for name in path_var_names
        ]

        path_match = URLUtils.match_routes(
            abstract_route=url_vars.abstract_route,
            real_route=url_vars.url.path
        )

        if path_match:
            for var_name, var_value in zip_longest(
                path_var_names, path_match.groups()
            ):
                if var_name is None or var_value is None:
                    raise ValueError(
                        "unmatched amount of variable names and it's values in"
                        " given abstract and real routes"
                    )

                url_vars.path_vars[var_name] = var_value

    @staticmethod
    def match_routes(
        *,
        abstract_route: str,
        real_route: str
    ) -> re.Match | None:
        """
        Returns match object of comparing real route against an abstract one.
        """
        regex_ready_abstract_route: str = re.sub(
            r"\\{\w+\\}",
            r"(\\w+)",
            re.escape(abstract_route)
        )
        return re.fullmatch(
            re.compile(regex_ready_abstract_route),
            real_route
        )

    @staticmethod
    def _set_query_vars(
        url_vars: URLVars
    ) -> None:
        query: str = url_vars.url.query
        query_elements: list[str] = query.split("&")
        for el in query_elements:
            if el == "":
                continue
            name, value = el.split("=")
            url_vars.query_vars[name] = value


class URLMethod(Enum):
    """
    All possible request methods.
    """
    Get = "get"
    Post = "post"
    Put = "put"
    Delete = "delete"
    Patch = "patch"
    Options = "options"
    Websocket = "websocket"



from enum import Enum
from typing import Any

from pykit import validation
from pytest import fixture

from orwynn.dto import ContainerDTO, UnitDTO
from orwynn.indication.indication import Indication
from orwynn.indication.indicator import Indicator
from orwynn.model.model import Model
from tests.std.text import Text


@fixture
def default_indication() -> Indication:
    mp: dict[str, Indicator] = {
        "type": Indicator.Type,
        "value": Indicator.Value
    }
    return Indication(mp)


def test_digest_default(default_indication: Indication):
    digested_mp: dict[str, Any] = default_indication.digest(
        Text(text="hello")
    )

    mp_type: str = validation.apply(digested_mp["type"], str)
    mp_value: dict = validation.apply(digested_mp["value"], dict)

    assert mp_type == "ok"
    Text.parse_obj(mp_value)


def test_recover_default(default_indication: Indication):
    recovering_mp: dict = {
        "type": "ok",
        "value": {
            "text": "hello"
        }
    }

    recovered_model: Text = default_indication.recover(Text, recovering_mp)
    assert type(recovered_model) is Text
    assert recovered_model.text == "hello"


def test_multiple_schemas():
    class Item(Model):
        name: str
        price: float

    i: Indication = Indication({
        "type": Indicator.Type,
        "value": Indicator.Value
    })

    assert i.gen_schema(Item) == i.gen_schema(Item)


def test_digest_enum(default_indication: Indication):
    class Color(Enum):
        RED = "red"
        GREEN = "green"

    class Priority(Enum):
        HIGH = 1
        LOW = 2

    class Item(Model):
        color: Color
        priority: Priority

        def __init__(self, **data: Any) -> None:
            super().__init__(**data)

    digested_mp: dict[str, Any] = default_indication.digest(
        Item(
            color="red",
            priority=1
        )
    )

    mp_value: dict = validation.apply(digested_mp["value"], dict)

    # Enum fields should be converted to values
    assert mp_value["color"] == "red"
    assert mp_value["priority"] == 1

    Item.parse_obj(mp_value)


def test_digest_error(default_indication: Indication):
    """
    Should digest error into API with external error code.
    """
    class ErrorCode(Enum):
        SIMPLE_CASE = "SIMPLE_CASE"
        ADVANCED_CASE = "ADVANCED_CASE"

    class SimpleError(Exception):
        Code = ErrorCode.SIMPLE_CASE

    class AdvancedError(Exception):
        Code = ErrorCode.ADVANCED_CASE

    class IntError(Exception):
        Code = 5

    class StrError(Exception):
        Code = "hello"

    data: dict

    data = default_indication.digest(SimpleError("hello world"))
    assert data["type"] == "error"
    assert data["value"]["message"] == "hello world"
    assert data["value"]["code"] == ErrorCode.SIMPLE_CASE.value

    data = default_indication.digest(AdvancedError("hello world"))
    assert data["type"] == "error"
    assert data["value"]["message"] == "hello world"
    assert data["value"]["code"] == ErrorCode.ADVANCED_CASE.value

    data = default_indication.digest(IntError("hello world"))
    assert data["type"] == "error"
    assert data["value"]["message"] == "hello world"
    assert data["value"]["code"] == 5

    data = default_indication.digest(StrError("hello world"))
    assert data["type"] == "error"
    assert data["value"]["message"] == "hello world"
    assert data["value"]["code"] == "hello"


def test_digest_polymorph(bare_boot):
    class ItemUDTO(UnitDTO):
        Code = "slimebones.orwynn.indication-testing.udto.item"
        price: float

    class SuperItemUDTO(ItemUDTO):
        Code = "slimebones.orwynn.indication-testing.udto.super-item"
        coolness: int

    class MiniItemUDTO(ItemUDTO):
        Code = "slimebones.orwynn.indication-testing.udto.mini-item"
        size: int

    class TinyItemUDTO(MiniItemUDTO):
        Code = "slimebones.orwynn.indication-testing.udto.tiny-item"
        is_really_small: bool

    class ItemCDTO(ContainerDTO):
        Base = ItemUDTO
        units: list[ItemUDTO]

    cdto = ItemCDTO(units=[
        ItemUDTO(id="1", price=10.5),
        SuperItemUDTO(id="2", price=50.3, coolness=100),
        MiniItemUDTO(id="3", price=2.1, size=3),
        TinyItemUDTO(id="4", price=1.5, size=1, is_really_small=True)
    ])

    api = cdto.api

    u = api["value"]["units"][0]
    assert u["id"] == "1"
    assert u["price"] == 10.5

    u = api["value"]["units"][1]
    assert u["id"] == "2"
    assert u["price"] == 50.3
    assert u["coolness"] == 100

    u = api["value"]["units"][2]
    assert u["id"] == "3"
    assert u["price"] == 2.1
    assert u["size"] == 3

    u = api["value"]["units"][3]
    assert u["id"] == "4"
    assert u["price"] == 1.5
    assert u["size"] == 1
    assert u["is_really_small"] is True



from typing import TYPE_CHECKING

from orwynn.worker import Worker

if TYPE_CHECKING:
    from orwynn.indication import Indication


class ApiIndicationOnlyProxy(Worker):
    def __init__(self, api_indication: "Indication") -> None:
        super().__init__()
        self.api_indication = api_indication



from orwynn.di.provider import Provider


class ProviderAlreadyInitializedForMapError(Exception):
    """If provider already initialized in some metamap."""
    def __init__(
        self, message: str = "", FailedProvider: type[Provider] | None = None
    ) -> None:
        if not message and FailedProvider:
            message = "{} already initialized for this map".format(
                FailedProvider
            )
        super().__init__(message)


class ProviderNotFoundInMapError(Exception):
    """If some requested provider not found in some metamap."""
    def __init__(
        self,
        message: str = "",
        P: type[Provider] | None = None
    ) -> None:
        if not message and P:
            message = f"{P} not found in map"
        super().__init__(message)


class ProviderKeywordAttributeError(Exception):
    """
    Providers cannot have keyword-only attributes, it's not logical for DI.
    """


class NoDependenciesForGivenProviderError(Exception):
    def __init__(
        self,
        message: str = "",
        P: type[Provider] | None = None
    ) -> None:
        if not message and P:
            message = f"{P} has no dependencies"
        super().__init__(message)




from typing import Callable, ItemsView

from pykit import validation

from orwynn.http.errors import EndpointNotFoundError
from orwynn.worker import Worker

from .endpoint import Endpoint


class EndpointContainer(Worker):
    """Collects endpoint specs to be used on routes registering."""
    def __init__(self) -> None:
        super().__init__()
        self.__spec_by_func: dict[Callable, Endpoint] = {}

    @property
    def items(self) -> ItemsView[Callable, Endpoint]:
        return self.__spec_by_func.items()

    def add(self, func: Callable, spec: Endpoint) -> None:
        validation.validate(func, Callable)
        validation.validate(spec, Endpoint)

        if func in self.__spec_by_func:
            raise ValueError(
                f"{func} already declared for proxy"
            )

        self.__spec_by_func[func] = spec

    def find_spec(self, func: Callable) -> Endpoint:
        validation.validate(func, Callable)

        try:
            return self.__spec_by_func[func]
        except KeyError as err:
            raise EndpointNotFoundError(
                f"{func} not found in proxy"
            ) from err



from pathlib import Path
from typing import TYPE_CHECKING

from orwynn.url import URLScheme
from orwynn.worker import Worker

if TYPE_CHECKING:
    from orwynn.apiversion import ApiVersion
    from orwynn.app import AppMode
    from orwynn.apprc import AppRc
    from orwynn.errorhandler.errorhandler import ErrorHandler
    from orwynn.indication.indication import Indication


class BootProxy(Worker):
    """Proxy data to prevent DI importing Boot worker directly (and avoid
    circular imports by this) to build BootConfig.
    """
    def __init__(
        self,
        *,
        root_dir: Path,
        mode: "AppMode",
        api_indication: "Indication",
        apprc: "AppRc",
        ErrorHandlers: set[type["ErrorHandler"]],
        global_http_route: str,
        global_websocket_route: str,
        api_version: "ApiVersion"
    ) -> None:
        super().__init__()
        self.__root_dir: Path = root_dir
        self.__mode: AppMode = mode
        self.__api_indication: Indication = api_indication
        self.__apprc: AppRc = apprc
        self.__ErrorHandlers: set[type["ErrorHandler"]] = \
            ErrorHandlers
        self.__global_http_route: str = global_http_route
        self.__global_websocket_route: str = global_websocket_route
        self.__api_version: ApiVersion = api_version

    @property
    def root_dir(self) -> Path:
        return self.__root_dir

    @property
    def mode(self) -> "AppMode":
        return self.__mode

    @property
    def api_indication(self) -> "Indication":
        return self.__api_indication

    @property
    def apprc(self) -> "AppRc":
        return self.__apprc

    @property
    def ErrorHandlers(self) -> set[type["ErrorHandler"]]:
        return self.__ErrorHandlers

    @property
    def global_http_route(self) -> str:
        return self.__global_http_route

    @property
    def global_websocket_route(self) -> str:
        return self.__global_websocket_route

    @property
    def api_version(self) -> "ApiVersion":
        return self.__api_version

    @property
    def boot_config_data(self) -> dict:
        return {
            "root_dir": self.__root_dir,
            "mode": self.__mode,
            "api_indication": self.__api_indication,
            "app_rc": self.__apprc,
            "ErrorHandlers": self.__ErrorHandlers
        }

    @property
    def data(self) -> dict:
        return dict(**self.boot_config_data, **{})

    def get_global_route_for_protocol(
        self,
        scheme: URLScheme
    ) -> str:
        if scheme is URLScheme.HTTP:
            return self.__global_http_route
        elif scheme is URLScheme.Websocket:
            return self.__global_websocket_route
        else:
            raise TypeError(
                f"unrecognized protocol {scheme}"
            )



from bson import ObjectId
from pykit import validation

from orwynn.mongo.document.errors import InvalidIdError
from orwynn.mongo.document.helpers import convert_to_object_id


def test_str():
    object_id: ObjectId = ObjectId()

    result: str | ObjectId = convert_to_object_id(str(object_id))

    assert type(result) is ObjectId
    assert result == object_id


def test_object_id():
    object_id: ObjectId = ObjectId()

    result: ObjectId = convert_to_object_id(object_id)

    assert type(result) is ObjectId
    assert result == object_id


def test_int():
    result: int | ObjectId = convert_to_object_id(1)

    assert type(result) is int
    assert result == 1


def test_dict():
    object_id: ObjectId = ObjectId()

    result: dict | ObjectId = convert_to_object_id({
        "hello": str(object_id)
    })

    assert type(result) is dict
    assert result["hello"] == object_id


def test_list():
    object_id_1: ObjectId = ObjectId()
    object_id_2: ObjectId = ObjectId()

    result: list | ObjectId = convert_to_object_id([
        str(object_id_1), str(object_id_2)
    ])

    assert type(result) is list
    assert len(result) == 2
    assert {str(x) for x in result} == {str(object_id_1), str(object_id_2)}


def test_invalid_id():
    validation.expect(
        convert_to_object_id,
        InvalidIdError,
        "hello"
    )



from enum import Enum


class Indicator(Enum):
    """List of indicators applicable to Schema's data representation.

    Values:
        Type:
            Indicates that field should contain any type representation of an
            object.
        Value:
            Indicates that field should contain a value of an object.
    """
    Type = "type"
    Value = "value"



import os
from pathlib import Path
from typing import Optional, Sequence

from pykit import validation
from pykit.errors import UnsupportedError
from pykit.env import EnvUtils
from pykit.errors import PleaseDefineError, StrExpectError
from sqlalchemy import Engine
from sqlalchemy import Table as SQLAlchemyTable
from sqlalchemy import create_engine, text
from sqlalchemy.orm import Session

from orwynn.app.mode import AppMode
from orwynn.database.database import Database
from orwynn.log.log import Log
from orwynn.proxy.boot import BootProxy
from orwynn.sql.config import SQLConfig
from orwynn.sql.constants import (PSQLStatementConstants)

from .enums import SQLDatabaseKind
from .table import Table


class SQL(Database):
    def __init__(
        self,
        config: SQLConfig
    ) -> None:
        super().__init__()
        self._config = config
        self._engine: Engine = self._create_engine()

    @property
    def engine(self) -> Engine:
        return self._engine

    @property
    def should_drop_sql(self) -> bool:
        if (self._config.should_drop_env_spec is None):
            raise PleaseDefineError(
                cannot_do="should drop sql environ retrieval",
                please_define="SQLConfig.should_drop_sql_env_spec"
            )

        return EnvUtils.get_bool(self._config.should_drop_env_spec)

    @property
    def session(self) -> Session:
        """Returns a new session."""
        return Session(self._engine)

    def create_tables(self, *tables: type[Table]) -> None:
        """Creates tables.

        Args:
            *tables:
                Tables to create. If no tables are given (or None), all
                reachable tables are created.
        """
        Table.metadata.create_all(
            self._engine,
            tables=self._collect_sqla_tables(*tables)
        )

    def drop_tables(self, *tables: type[Table]) -> None:
        """Drops tables.

        Args:
            *tables:
                Tables to drop. If no tables are given (or None), all
                reachable tables are dropped.
        """
        Table.metadata.drop_all(
            self._engine,
            tables=self._collect_sqla_tables(*tables)
        )

    def recreate_public_schema_cascade(
        self
    ) -> None:
        if (self._config.database_kind is SQLDatabaseKind.SQLite):
            raise UnsupportedError(
                value="sqlite database for cascade recreating"
            )

        if (
            # lock drop in prod mode for error-protection
            BootProxy.ie().mode != AppMode.PROD
        ):
            Log.info("recreating sql database (public cascade)")
            # use raw sql since i haven't found a quick way to drop cascade
            # using python objects
            statement: str = (
                PSQLStatementConstants.RecreatePublicSchemaCascade
            )

            with self.session as s:
                s.execute(text(statement))
                s.commit()

    def _collect_sqla_tables(
        self,
        *tables: type[Table]
    ) -> Optional[Sequence[SQLAlchemyTable]]:
        sqla_tables: Optional[Sequence[SQLAlchemyTable]] = None
        if tables:
            sqla_tables = []
            for table in tables:
                sqla_tables.append(table.__table__)  # type: ignore
        return sqla_tables

    def _create_engine(self) -> Engine:
        url: str
        connect_args: dict = {}
        final_kwargs: dict = {}

        match self._config.database_kind:
            case SQLDatabaseKind.PostgreSQL:
                url = \
                    f"postgresql://{self._config.database_user}" \
                    + f":{self._config.database_password}" \
                    + f"@{self._config.database_host}" \
                    + f":{self._config.database_port}" \
                    + f"/{self._config.database_name}"
            case SQLDatabaseKind.SQLite:
                database_path: str = validation.apply(
                    self._config.database_path,
                    str
                )

                # Create directories for the path if they are not exist.
                # https://stackoverflow.com/a/273227/14748231
                Path(database_path).parent.mkdir(parents=True, exist_ok=True)

                # Note that here we add additional slash since in either
                # memory ":memory:", relative "var/app.db" or absolute
                # "/tmp/app.db" cases we have to prepend slash to comply with
                # SQLite standard.
                url = f"sqlite:///{database_path}"

                # Make SQLite know that more than one thread could interact
                # with the database for the same request.
                # See https://fastapi.tiangolo.com/tutorial/sql-databases/#note
                connect_args.update({"check_same_thread": False})
            case _:
                raise TypeError(
                    f"unknown database {self._config.database_kind}"
                )

        final_kwargs["connect_args"] = connect_args
        final_kwargs["poolclass"] = self._config.real_poolclass
        if self._config.pool_size:
            final_kwargs["pool_size"] = self._config.pool_size

        return create_engine(
            url,
            **final_kwargs
        )



class CircularDependencyError(Exception):
    pass


class FrameworkServiceModuleReferenceError(Exception):
    pass


class EmptyRouteError(Exception):
    def __init__(self, message: str = "") -> None:
        if not message:
            message = "route cannot be empty"
        super().__init__(message)



from orwynn.error.errors import (ExceptionAlreadyHandledError,
                                      MalfunctionError)
from orwynn.errorhandler import ErrorHandler
from orwynn.http import DEFAULT_HTTP_ERROR_HANDLERS
from orwynn.url import URLScheme
from orwynn.websocket import DEFAULT_WEBSOCKET_EXCEPTION_HANDLERS


class ErrorHandlerManager:
    """
    Controls the error flow setup for the system.

    Attributes:
        register_by_protocol:
            Dictionary of register functions for each protocol.
    """
    def get_populated_handlers_by_protocol(
        self,
        exception_handlers: set[ErrorHandler]
    ) -> dict[URLScheme, set[ErrorHandler]]:
        """
        Forms a set of handlers populated with default ones if required.

        Args:
            exception_handlers:
                Set of error handlers to register.

        Returns:
            Populated set of exception handlers by their protocol.
        """
        handlers_by_protocol: dict[URLScheme, set[ErrorHandler]] = {}

        # Populate and register handlers separately for each protocol
        for scheme in URLScheme:
            if (
                scheme is not URLScheme.HTTP
                and scheme is not URLScheme.Websocket
            ):
                continue

            populated_handlers: set[ErrorHandler] = \
                self.__populate_handlers(
                    self.__get_handlers_for_protocol(
                        scheme,
                        exception_handlers
                    ),
                    scheme
                )

            if scheme in handlers_by_protocol:
                raise ValueError(
                    f"protocol {scheme} handled twice"
                )
            handlers_by_protocol[scheme] = populated_handlers

        return handlers_by_protocol

    def __get_handlers_for_protocol(
        self,
        protocol: URLScheme,
        error_handlers: set[ErrorHandler]
    ) -> set[ErrorHandler]:
        """
        Collects all handlers for the given protocol.
        """
        final_set: set[ErrorHandler] = set()

        for eh in error_handlers:
            if eh.Protocol is protocol:
                final_set.add(eh)

        return final_set

    def __populate_handlers(
        self,
        error_handlers: set[ErrorHandler],
        protocol: URLScheme
    ) -> set[ErrorHandler]:
        """
        Traverses handlers to find handled Python-builtin exceptions.

        Returns:
            Set of populated handlers.
        """
        populated_handlers: set[ErrorHandler] = set()
        __HandledExceptions: set[type[Exception]] = set()

        for eh in error_handlers:
            if eh.E is None:
                raise MalfunctionError(
                    f"handler {eh} attribute E cannot be None"
                )
            else:
                if eh.E in __HandledExceptions:
                    raise ExceptionAlreadyHandledError(
                        f"exception {eh.E} is already handled"
                    )
                elif issubclass(eh.E, Exception):
                    __HandledExceptions.add(eh.E)
                else:
                    raise MalfunctionError(
                        f"unrecognized error {eh.E}"
                    )

            populated_handlers.add(eh)

        # Add default handlers for errors for which the custom's handlers were
        # not added
        DEFAULT_HANDLERS: set[type[ErrorHandler]]
        match protocol:
            case URLScheme.HTTP:
                DEFAULT_HANDLERS = DEFAULT_HTTP_ERROR_HANDLERS
            case URLScheme.Websocket:
                DEFAULT_HANDLERS = DEFAULT_WEBSOCKET_EXCEPTION_HANDLERS
            case _:
                raise TypeError(
                    f"unrecognized protocol {protocol}"
                )

        for GenericDefaultErrorHandler in DEFAULT_HANDLERS:
            if GenericDefaultErrorHandler.E not in __HandledExceptions:
                populated_handlers.add(
                    GenericDefaultErrorHandler()
                )

        return populated_handlers



import inspect
from types import NoneType
from typing import TYPE_CHECKING, Any, Callable, Optional, Self, TypeVar
from orwynn.apiversion import ApiVersion
from orwynn.model.model import Model
from orwynn.proxy.boot import BootProxy

from orwynn.testing.embeddedclient import EmbeddedTestClient
from pykit import validation
from pykit.validation import validate
from orwynn.url import URLScheme, URLUtils

if TYPE_CHECKING:
    from orwynn.http import TestHttpResponse

# If ever you get to Python3.12, see if PEP 696 introduced, then apply
# but for now it is in the next form
_JsonifyExpectedType = TypeVar("_JsonifyExpectedType")


class _FinalizedRequestData(Model):
    route: str
    kwargs: dict


class Client:
    """
    Operates with HTTP client requests for test purposes.
    """
    def __init__(
        self,
        embedded_client: EmbeddedTestClient,
        *,
        binded_headers: dict[str, str] | None = None
    ) -> None:
        validation.validate(embedded_client, EmbeddedTestClient)
        self._embedded_client: EmbeddedTestClient = embedded_client

        if binded_headers is None:
            binded_headers = {}
        # Do copy here again for an additional safety, although headers copying
        # is already done at bind_headers() method.
        self._binded_headers: dict[str, str] = binded_headers.copy()

    @property
    def binded_headers(self) -> dict:
        return self._binded_headers

    def bind_headers(
        self,
        headers: dict[str, str]
    ) -> Self:
        validation.validate_dict(headers, (str, str))

        # Accumulate headers from this client to the new one.
        #
        # Also copy binded headers to not stack them after multiple
        # bind_headers() calls.
        final_headers: dict[str, str] = self._binded_headers.copy()
        final_headers.update(headers)

        return self.__class__(
            self._embedded_client,
            binded_headers=final_headers
        )

    def get_jsonify(
        self,
        url: str,
        asserted_status_code: int | None = None,
        *,
        expected_type: type[_JsonifyExpectedType] = dict,
        **kwargs
    ) -> _JsonifyExpectedType:
        r: "TestHttpResponse" = self.get(url, asserted_status_code, **kwargs)
        data: Any = r.json()
        validate(data, expected_type)
        return data

    def post_jsonify(
        self,
        url: str,
        asserted_status_code: int | None = None,
        *,
        expected_type: type[_JsonifyExpectedType] = dict,
        **kwargs
    ) -> _JsonifyExpectedType:
        r: "TestHttpResponse" = self.post(url, asserted_status_code, **kwargs)
        data: Any = r.json()
        validate(data, expected_type)
        return data

    def delete_jsonify(
        self,
        url: str,
        asserted_status_code: int | None = None,
        *,
        expected_type: type[_JsonifyExpectedType] = dict,
        **kwargs
    ) -> _JsonifyExpectedType:
        r: "TestHttpResponse" = self.delete(
            url, asserted_status_code, **kwargs
        )
        data: Any = r.json()
        validate(data, expected_type)
        return data

    def put_jsonify(
        self,
        url: str,
        asserted_status_code: int | None = None,
        *,
        expected_type: type[_JsonifyExpectedType] = dict,
        **kwargs
    ) -> _JsonifyExpectedType:
        r: "TestHttpResponse" = self.put(url, asserted_status_code, **kwargs)
        data: Any = r.json()
        validate(data, expected_type)
        return data

    def patch_jsonify(
        self,
        url: str,
        asserted_status_code: int | None = None,
        *,
        expected_type: type[_JsonifyExpectedType] = dict,
        **kwargs
    ) -> _JsonifyExpectedType:
        r: "TestHttpResponse" = self.patch(url, asserted_status_code, **kwargs)
        data: Any = r.json()
        validate(data, expected_type)
        return data

    def options_jsonify(
        self,
        url: str,
        asserted_status_code: int | None = None,
        *,
        expected_type: type[_JsonifyExpectedType] = dict,
        **kwargs
    ) -> _JsonifyExpectedType:
        r: "TestHttpResponse" = self.options(
            url, asserted_status_code, **kwargs
        )
        data: Any = r.json()
        validate(data, expected_type)
        return data

    def get(
        self,
        url: str,
        asserted_status_code: int | None = None,
        **kwargs
    ) -> "TestHttpResponse":
        # Join method name and function to call inner resolver
        # inspect.stack() is for resolving self method name, ref:
        #   https://stackoverflow.com/a/5067654
        return self._get_test_response(
            inspect.stack(), url, asserted_status_code, **kwargs)

    def post(
        self,
        url: str,
        asserted_status_code: int | None = None,
        **kwargs
    ) -> "TestHttpResponse":
        return self._get_test_response(
            inspect.stack(), url, asserted_status_code, **kwargs)

    def delete(
        self,
        url: str,
        asserted_status_code: int | None = None,
        **kwargs
    ) -> "TestHttpResponse":
        return self._get_test_response(
            inspect.stack(), url, asserted_status_code, **kwargs)

    def put(
        self,
        url: str,
        asserted_status_code: int | None = None,
        **kwargs
    ) -> "TestHttpResponse":
        return self._get_test_response(
            inspect.stack(), url, asserted_status_code, **kwargs)

    def patch(
        self,
        url: str,
        asserted_status_code: int | None = None,
        **kwargs
    ) -> "TestHttpResponse":
        return self._get_test_response(
            inspect.stack(), url, asserted_status_code, **kwargs)

    def options(
        self,
        url: str,
        asserted_status_code: int | None = None,
        **kwargs
    ) -> "TestHttpResponse":
        return self._get_test_response(
            inspect.stack(), url, asserted_status_code, **kwargs)

    def _get_test_response(
        self,
        stack: list[inspect.FrameInfo],
        url: str,
        asserted_status_code: int | None,
        **kwargs
    ) -> "TestHttpResponse":
        request: str = " ".join([stack[0][3], url])
        return self._resolve_request(request, asserted_status_code, **kwargs)

    def _resolve_request(
        self,
        request: str,
        asserted_status_code: int | None,
        **request_kwargs
    ) -> "TestHttpResponse":
        response: "TestHttpResponse"
        test_client_method: Callable
        method: str
        route: str

        validate(request, str)
        validate(asserted_status_code, [int, NoneType])

        # Request example: "get /users/1"
        method, route = request.split(" ")

        # Also can accept uppercase "GET ..."
        method = method.lower()

        match method:
            case "get":
                test_client_method = self._embedded_client.get
            case "post":
                test_client_method = self._embedded_client.post
            case "delete":
                test_client_method = self._embedded_client.delete
            case "put":
                test_client_method = self._embedded_client.put
            case "patch":
                test_client_method = self._embedded_client.patch
            case "options":
                test_client_method = self._embedded_client.options
            case _:
                raise ValueError(f"Method {method} is not supported")

        # Get finalized data
        finalized: _FinalizedRequestData = self._process_request_data(
            route=route,
            request_kwargs=request_kwargs,
            protocol=URLScheme.HTTP
        )

        # Make a request
        response: "TestHttpResponse" = test_client_method(
            finalized.route,
            **finalized.kwargs
        )

        if asserted_status_code is not None:
            assert \
                response.status_code == asserted_status_code, \
                f"response status code {response.status_code}" \
                f" != asserted status code {asserted_status_code};" \
                f" response content is {response.content}"

        return response

    def _process_request_data(
        self,
        *,
        route: str,
        request_kwargs: dict,
        protocol: URLScheme
    ) -> _FinalizedRequestData:
        """
        Processes a given route and request kwargs and returns finalized
        versions of them.
        """
        request_kwargs = request_kwargs.copy()

        # Add binded headers
        request_kwargs.setdefault("headers", {})
        request_kwargs["headers"].update(self._binded_headers)

        # Craft the final route
        is_global_route_used: bool = validation.apply(
            request_kwargs.get("is_global_route_used", True),
            bool
        )
        api_version: int | None = request_kwargs.get("api_version", None)
        validation.validate(api_version, [int, NoneType])
        final_route: str = self._get_final_route(
            route,
            is_global_route_used=is_global_route_used,
            api_version=api_version,
            protocol=protocol
        )
        # Delete custom keys to not confuse a client's method
        try:
            del request_kwargs["is_global_route_used"]
        except KeyError:
            pass
        try:
            del request_kwargs["api_version"]
        except KeyError:
            pass

        return _FinalizedRequestData(
            route=final_route,
            kwargs=request_kwargs
        )

    def _get_final_route(
        self,
        route: str,
        *,
        is_global_route_used: bool,
        api_version: int | None,
        protocol: URLScheme
    ) -> str:
        if not is_global_route_used and api_version is not None:
            raise ValueError(
                f"the global route is disabled and api version is also passed"
                " which doesn't make sense"
            )
        elif not is_global_route_used:
            return route

        api_version_obj: ApiVersion = BootProxy.ie().api_version

        global_route: str = BootProxy.ie().get_global_route_for_protocol(
            protocol
        )

        final_api_version: int

        if api_version:
            final_api_version = api_version
        else:
            final_api_version = api_version_obj.latest

        final_global_route: str
        try:
            final_global_route = api_version_obj.apply_version_to_route(
                global_route,
                final_api_version
            )
        except ValueError:
            # No {version} format block
            final_global_route = global_route

        return URLUtils.join_routes(
            final_global_route,
            route
        )

    def websocket(
        self,
        route: str,
        **request_kwargs
    ) -> Any:
        finalized: _FinalizedRequestData = self._process_request_data(
            route=route,
            request_kwargs=request_kwargs,
            protocol=URLScheme.Websocket
        )

        return self._embedded_client.websocket_connect(
            finalized.route,
            **finalized.kwargs
        )



from enum import Enum

from pykit import validation

from orwynn.mongo import MongoUtils
from orwynn.mongo.errors import MongoTypeConversionError


class StringEnum(Enum):
    Red = "red"
    Green = "green"


class StringObj:
    @property
    def mongovalue(self) -> str:
        return "hello"


class DictObj:
    @property
    def mongovalue(self) -> dict:
        return {
            1: "hello",
            2: "world",
        }


class SetObj:
    @property
    def mongovalue(self) -> set:
        return {1, 2, 3}


def test_already_compatible():
    """
    Should work correctly for already compatible types.
    """
    assert MongoUtils.convert_compatible("hello") == "hello"
    assert MongoUtils.convert_compatible(2) == 2
    assert MongoUtils.convert_compatible(2.3) == 2.3
    assert MongoUtils.convert_compatible(True) is True
    assert MongoUtils.convert_compatible([1, 2, 3]) == [1, 2, 3]
    assert MongoUtils.convert_compatible({
        1: "hello",
        2: "world",
    }) == {
        1: "hello",
        2: "world",
    }
    assert MongoUtils.convert_compatible(None) is None


def test_containers():
    """
    Should work correctly for containers, like list or dict, which also may
    contain sub-containers.
    """
    mock_dict: dict = {
        1: {
            1: "hello",
            2: "world",
        },
        2: [
            {
                1: "hello",
                2: "world",
            },
            {
                1: "here",
                2: "we",
                3: "go",
                4: ["again", "!!"],
            },
        ],
    }

    assert MongoUtils.convert_compatible(mock_dict) == mock_dict


def test_enum_string():
    """
    Should convert enum correctly.
    """
    assert MongoUtils.convert_compatible(StringEnum.Red) == "red"


def test_mongovalue_string():
    """
    Should convert objects with attribute `mongovalue`.
    """
    assert MongoUtils.convert_compatible(StringObj()) == "hello"


def test_mongovalue_dict():
    """
    Should convert objects with attribute `mongovalue` which returns
    dict.
    """
    assert MongoUtils.convert_compatible(DictObj()) == {
        1: "hello",
        2: "world",
    }


def test_mongovalue_set():
    """
    Should raise an error for set returning mongovalue attribute.
    """
    validation.expect(
        MongoUtils.convert_compatible,
        MongoTypeConversionError,
        SetObj(),
    )



from orwynn.errorhandler import ErrorHandler
from orwynn.http.errorhandler.middleware import ErrorHandlerHttpMiddleware
from orwynn.http.middleware.builtinmiddleware import BuiltinHttpMiddleware

from .context.middleware.builtin import ContextBuiltinMiddleware
from .context.middleware.contextbuiltin import (
    HttpRequestContextBuiltinMiddleware,
)
from .errorhandler.default import (
    DefaultErrorHandler,
    DefaultHttpErrorHandler,
    DefaultRequestValidationErrorHandler,
)

DEFAULT_HTTP_ERROR_HANDLERS: set[type[ErrorHandler]] = {
    DefaultErrorHandler,
    DefaultHttpErrorHandler,
    DefaultRequestValidationErrorHandler
}
"""
List of error handlers applied by default.
"""

# Order matters, the lowest index is initialized first.
BUILTIN_HTTP_MIDDLEWARE: list[type[BuiltinHttpMiddleware]] = [
    ErrorHandlerHttpMiddleware,
    ContextBuiltinMiddleware,
    HttpRequestContextBuiltinMiddleware
]



from typing import Self

from orwynn.model.model import Model
from orwynn.url import URLScheme


class Address(Model):
    """
    Representation of URL.

    Attributes:
        host:
            Host of the address.
        top_domain(optional):
            Top-level domain. Defaults to host name in case of ip-like
            hostname (such as `127.0.0.1`).
        second_domain(optional):
            Second-level domain. Defaults to None.
        subdomain(optional):
            Subdomain. Defaults to None.
        port(optional):
            Port of the address. Defaults to None.
        protocol(optional):
            Protocol of the address. Defaults to None.
        path(optional):
            Route of the address. Defaults to None.
        query(optional):
            Query arguments of the address. Query's values are not converted
            to basic types, all of them are strings or list of strings - for
            multiple occured keys, all occured values are added to the list
            for the key. Defaults to None.
        username(optional):
            Username of the address. Defaults to None.
        password(optional):
            Password of the address. Defaults to None.
        fragment(optional):
            Fragment of the address. Defaults to None.
    """
    host: str
    top_domain: str | None = None
    second_domain: str | None = None
    subdomain: str | None = None
    scheme: URLScheme | None = None
    port: int | None = None
    path: str | None = None
    query: dict[str, str | list[str]] | None = None
    username: str | None = None
    password: str | None = None
    fragment: str | None = None

    @classmethod
    def parse(cls, raw_address: str) -> Self:
        """
        Parses a raw address into the model.

        Args:
            raw_address:
                Address to parse.

        Returns:
            Address model.

        Raises:
            InvalidAddressError:
                The given raw address does not match address's regex.
        """
        raise NotImplementedError()



from orwynn.config import Config
from orwynn.di.errors import NotProviderError, ProviderAvailabilityError
from orwynn.di.isprovider import is_provider
from orwynn.di.object import DiObject
from orwynn.di.provider import Provider
from orwynn.module.module import Module
from orwynn.service.framework import FrameworkService


def check_availability(
    Requestor: type[DiObject],
    Target: type[Provider],
    requestor_module: Module | None
) -> Module | None:
    """
    Checks if the Target is available to Requestor.

    Note, that this function is not checking provider priorities. So it's
    possible to see positive output here for lower-priority provider
    requesting higher-priority one, which in general logic shouldn't
    be possible.

    Resolves cases with frameworks services right away:
    - If P2 is a FrameworkService, we can be assured that it is available
        for any other requested Provider, except for cases where the requested
        Provider is a Config, because the Config is the only class that has
        a higher priority than the FrameworkService.

    Args:
        Requestor:
            Which object is searching.
        Target:
            Provider search is made for.
        requestor_module:
            Module of the Requestor to start searching from.
            Can be set to None to search within the Framework's scope.

    Returns:
        module:
            The module where Target is located
        None:
            If the Target is a FrameworkService, 100% available
        None:
            If the Requestor is a FrameworkService and the Target is any
            Config because we 100% sure that Config will be available.

    Raises:
        ProviderAvailabilityError:
            The Target provider is not available for the Requestor.
    """
    is_error: bool
    res: Module | None = None

    # The Target should be provider
    if not is_provider(Target):
        raise NotProviderError(
            f"object {Target} is not a provider"
        )
    elif (
        issubclass(Requestor, FrameworkService) and issubclass(Target, Config)
        or (
            issubclass(Target, FrameworkService)
            and not issubclass(Requestor, Config)
        )
    ):
        is_error = False
    elif requestor_module is not None:
        found_module = __search_matching_provider(
            target_module=requestor_module,
            TargetProvider=Target
        )
        if found_module is None:
            is_error = True
        else:
            is_error = False
            res = found_module
    else:
        raise ProviderAvailabilityError(
            f"not logical to check an availability between {Requestor} ->"
            f" {Target} in the framework's scope"
        )

    if is_error:
        raise ProviderAvailabilityError(Provider1=Requestor, DiObject=Target)
    else:
        return res


def __search_matching_provider(
    target_module: Module,
    TargetProvider: type[Provider]
) -> Module | None:
    """
    Searches for the provider in module.

    Returns:
        The module where the provider was found, or None, it it wasn't.
    """

    res: Module | None = None
    is_found: bool = False

    for P in target_module.Providers:
        if P is TargetProvider:
            # Available within the same module
            is_found = True
            res = target_module

    if not is_found:
        # Not available in the same module, check the exports of the imported
        # ones
        for m in target_module._imports:
            for P in m._exports:
                if P is TargetProvider:
                    # Available in the exports of the imported module
                    is_found = True
                    res = m

    return res



from orwynn.di.constants import BUILTIN_PROVIDERS


def is_provider(Class: type) -> bool:
    """Checks if given class is a Provider.

    Args:
        Class:
            Any class to check to.

    Returns:
        Flag signifies if given class is a Provider.
    """
    return issubclass(Class, tuple(BUILTIN_PROVIDERS))



from orwynn.model.model import Model


class Cors(Model):
    """Represents configuration for special CORS middleware."""
    is_enabled: bool = True
    allow_origins: list[str] | None = None
    allow_origin_regex: str | None = None
    allow_methods: list[str] | None = None
    allow_headers: list[str] | None = None
    allow_credentials: bool = False
    expose_headers: list[str] | None = None
    max_age: int = 600



import pytest

from orwynn.boot import Boot
from orwynn.bootscript import Bootscript, CallTime
from orwynn.di.di import Di
from orwynn.sql import SQL, Table
from orwynn.sql.search import TableSearch
from orwynn.sql.shd import SHD
from orwynn.sql.table import Mapped
from orwynn.sql.utils import SQLUtils


class S1Item(Table):
    name: Mapped[str]
    price: Mapped[float]


@pytest.fixture
def create_tables_bootscript() -> Bootscript:
    return Bootscript(
        func=SQLUtils.create_tables,
        call_time=CallTime.AFTER_ALL,
    )


@pytest.fixture
def s1_table_search(bare_boot: Boot) -> TableSearch:
    sql: SQL = Di.ie().find("SQL")
    return TableSearch(
        ids=["a", "b", "c"],
        shd=SHD.new(sql),
    )



import pytest_asyncio

from orwynn import mongo
from orwynn.boot.boot import Boot
from orwynn.module.module import Module


@pytest_asyncio.fixture
async def mongo_boot() -> Boot:
    return await Boot.create(
        Module(route="/", imports=[mongo.module]),
        apprc={
            "prod": {
                "Mongo": {
                    "url": "mongodb://localhost:9006",
                    "database_name": "orwynn-test"
                }
            }
        }
    )



from enum import Enum

from pykit import validation

from orwynn.error.code import get_error_code
from orwynn.indication.indication import Indication
from orwynn.indication.indicator import Indicator


def test_int():
    class MyError(Exception):
        Code = 1

    assert get_error_code(MyError()) == 1


def test_str():
    class MyError(Exception):
        Code = "hello"

    assert get_error_code(MyError()) == "hello"


def test_enum():
    class ErrorCode(Enum):
        RED = "RED"

    class MyError(Exception):
        Code = ErrorCode.RED

    assert get_error_code(MyError()) == ErrorCode.RED


def test_wrong_type():
    class MyError(Exception):
        Code = 5.55

    validation.expect(
        get_error_code,
        validation.ValidationError,
        MyError()
    )


def test_no_attribute():
    class MyError(Exception):
        pass

    validation.expect(
        get_error_code,
        AttributeError,
        MyError()
    )

def test_default_indication_type():
    class E(Exception):
        pass

    i: Indication = Indication({
        "type": Indicator.Type,
        "value": Indicator.Value
    })

    assert i.digest(E("whatever"))["type"] == "error"



from orwynn.model.model import Model

from .dispatchfunc import DispatchWebsocketFn
from .genericfunc import GenericWebsocketFn


class WebsocketHandler(Model):
    func: GenericWebsocketFn
    route: str


class DispatchWebsocketHandler(WebsocketHandler):
    func: DispatchWebsocketFn
    # Middleware is responding to all routes, but inside own dispatch function
    # decides whether to handle it or not.
    route: str = "*"



from orwynn.indication.indication import Indication
from orwynn.indication.indicator import Indicator

default_api_indication: Indication = Indication(
    {
        "type": Indicator.Type,
        "value": Indicator.Value
    }
)




from bson import ObjectId
from bson.errors import InvalidId
from pykit.types import T

from orwynn.mongo.document.errors import InvalidIdError


def convert_to_object_id(obj: T) -> T | ObjectId:
    """
    Converts an object to ObjectId compliant.

    If the object is:
    - string: It is passed directly to ObjectId()
    - dict: All values are recursively converted using this method.
    - list: All items are recursively converted using this method.
    - other types: Nothing will be done.

    Returns:
        ObjectId-compliant representation of the given object.
    """
    result: T | ObjectId

    if type(obj) is str:
        try:
            result = ObjectId(obj)
        except InvalidId as error:
            raise InvalidIdError(
                invalid_id=obj
            ) from error
    elif type(obj) is dict:
        result = type(obj)()
        for k, v in obj.items():
            result[k] = convert_to_object_id(v)
    elif type(obj) is list:
        result = type(obj)([convert_to_object_id(x) for x in obj])
    else:
        result = obj

    return result



import pytest
from pydantic import RootModel
from pykit import validation
from pykit.map import MapUtils

from orwynn.app import AppMode
from orwynn.apprc.apprc import AppRc
from orwynn.boot import Boot
from orwynn.config import Config
from orwynn.di.di import Di
from orwynn.module.module import Module


class Menu(RootModel):
    root: dict[str, float]


class BurgerShotConfig(Config):
    location: str
    employees: int
    menu: Menu


@pytest.fixture
def raw_apprc() -> AppRc:
    return {
        "prod": {
            "BurgerShot": {
                "location": "Vinewood",
                "employees": 15,
                "menu": {
                    "standard_burger": 1.25,
                    "cola": 1.5
                }
            }
        },
        "dev": {
            "BurgerShot": {
                "employees": 25,
                "menu": {
                    "cola": 1.8,
                    "pizza": 4.1,
                    "fried_chicken": 3.5
                }
            }
        },
        "test": {
            "BurgerShot": {
                "location": "Jefferson",
                "menu": {
                    "standard_burger": 1.1,
                    "fried_chicken": 5.0
                }
            }
        }
    }


@pytest.mark.asyncio
async def test_prod(raw_apprc: AppRc):
    await Boot.create(
        root_module=Module("/", Providers=[BurgerShotConfig]),
        apprc=raw_apprc,
        mode=AppMode.PROD
    )

    assert validation.apply(
        Di.ie().find("BurgerShotConfig"),
        BurgerShotConfig
    ).model_dump() == MapUtils.find("prod.BurgerShot", raw_apprc)


@pytest.mark.asyncio
async def test_dev(raw_apprc: AppRc):
    await Boot.create(
        root_module=Module("/", Providers=[BurgerShotConfig]),
        apprc=raw_apprc,
        mode=AppMode.DEV
    )

    assert validation.apply(
        Di.ie().find("BurgerShotConfig"),
        BurgerShotConfig
    ).model_dump() == {
        "location": "Vinewood",
        "employees": 25,
        "menu": {
            "standard_burger": 1.25,
            "cola": 1.8,
            "pizza": 4.1,
            "fried_chicken": 3.5
        }
    }


@pytest.mark.asyncio
async def test_test(raw_apprc: AppRc):
    await Boot.create(
        root_module=Module("/", Providers=[BurgerShotConfig]),
        apprc=raw_apprc,
        mode=AppMode.TEST
    )

    assert validation.apply(
        Di.ie().find("BurgerShotConfig"),
        BurgerShotConfig
    ).dict() == {
        "location": "Jefferson",
        "employees": 25,
        "menu": {
            "standard_burger": 1.1,
            "cola": 1.8,
            "pizza": 4.1,
            "fried_chicken": 5.0
        }
    }



from enum import Enum


class IndicationType(Enum):
    """
    Possible values of field "type".

    Values:
        OK:
            Everything is correct.
        ERROR:
            Some error is happened either by client's problem or by server's
            issue. Here goes also all errors connected with business-logic
            itself.
    """
    Ok = "ok"
    # TODO(ryzhovalex):
    #   In future, consider moving all non-BL related problems
    #   to another type, e.g. "http-error".
    Error = "error"



import pytest
from pykit import validation

from orwynn.boot.boot import Boot
from orwynn.context.errors import UndefinedStorageError
from orwynn.http import (
    Endpoint,
    HttpController,
    HttpRequestContextId,
    LogMiddleware,
)
from orwynn.module.module import Module


@pytest.mark.asyncio
async def test_basic():
    """Request id should be fetchable from context within request-response
    cycle and unfetchable outside this cycle.
    """
    class C1(HttpController):
        Route = "/"
        Endpoints = [Endpoint(method="get")]

        def get(self) -> dict:
            return {"request_id": HttpRequestContextId().get()}

    boot: Boot = await Boot.create(
        Module(
            "/",
            Controllers=[C1],
            Middleware=[LogMiddleware]
        )
    )

    validation.expect(
        HttpRequestContextId().get,
        UndefinedStorageError
    )

    data: dict = boot.app.client.get_jsonify("/", 200)
    assert type(data["request_id"]) is str and data["request_id"] != ""






# TODO(ryzhovalex): in development
# 0
# @pytest.mark.parametrize(
#     "raw,expected",
#     [
#         (
#             "http://github.com",
#             Address(
#                 host="github.com",
#                 top_domain="com",
#                 second_domain="github",
#                 scheme=Scheme.HTTP
#             )
#         ),

#         (
#             "https://github.com/ryzhovalex/staze/blob/master/staze/core/cli/cli.py#L67",
#             Address(
#                 host="github.com",
#                 top_domain="com",
#                 second_domain="github",
#                 scheme=Scheme.HTTPS,
#             )
#         ),

#         (
#             "stackoverflow.com/questions/39625696/regex-url-capturing-group?hello=1&ids=22",
#             Address()
#         ),

#         (
#             "https://jsc-rpe-almaz.kaiten.ru/space/505050",
#             Address()
#         ),

#         (
#             "10.2.16.4:5000/hello",
#             Address()
#         ),
#     ]
# )
# def test_parse(raw: str, expected: Address):
#     pass



import pytest

from orwynn import Module
from orwynn.boot import Boot
from orwynn.http import Endpoint, HttpController
from orwynn.http.log.middleware import LogMiddleware
from orwynn.http.responses import RedirectHttpResponse, TestHttpResponse


@pytest.mark.asyncio
async def test_redirect():
    class LocalController(HttpController):
        Route = "/"
        Endpoints = [
            Endpoint(
                method="get"
            )
        ]

        def get(self) -> RedirectHttpResponse:
            return RedirectHttpResponse("https://google.com")


    boot: Boot = await Boot.create(
        Module("/", Controllers=[LocalController]),
        global_middleware={
            LogMiddleware: ["*"]
        }
    )  # type: ignore

    response: TestHttpResponse = boot.app.client.get(
        "/", 307, follow_redirects=False
    )

    assert response.headers.get("location", "") == "https://google.com"



import pytest
from pykit import validation

from orwynn.boot import Boot
from orwynn.bootscript.bootscript import Bootscript
from orwynn.bootscript.calltime import CallTime
from orwynn.di.di import Di
from orwynn.module import Module
from orwynn.service import Service


class SomeService(Service):
    def __init__(self) -> None:
        super().__init__()
        self.some_var: int = 0


def func(some_service: SomeService) -> None:
    some_service.some_var = 1


async def async_func(some_service: SomeService) -> None:
    some_service.some_var = 5


@pytest.mark.asyncio
async def test_basic():
    await Boot.create(
        Module(Providers=[SomeService]),
        bootscripts=[
            Bootscript(
                func=func,
                call_time=CallTime.AFTER_ALL
            )
        ]
    )

    some_service: SomeService = validation.apply(
        Di.ie().find("SomeService"),
        SomeService
    )

    assert some_service.some_var == 1


@pytest.mark.asyncio
async def test_async():
    await Boot.create(
        Module(Providers=[SomeService]),
        bootscripts=[
            Bootscript(
                func=func,
                call_time=CallTime.AFTER_ALL
            )
        ]
    )

    some_service: SomeService = validation.apply(
        Di.ie().find("SomeService"),
        SomeService
    )

    assert some_service.some_var == 1



from pykit import validation

from orwynn.http import TestHttpResponse
from orwynn.mapping.errors import CustomUseOfMappingReservedFieldError
from orwynn.mongo.document import Document
from orwynn.mongo.errors import DuplicateKeyError
from orwynn.proxy.boot import BootProxy
from orwynn.testing.client import Client
from tests.std.user import User


def test_user_create(std_mongo_boot, std_http: Client):
    r: TestHttpResponse = std_http.post(
        "/users",
        200,
        json={
            "name": "Mark Watney"
        }
    )
    created_user: User = User.recover(r.json())
    User.find_one({"id": created_user.id})


def test_reserved_mapping_field(std_mongo_boot, std_http: Client):
    class M(Document):
        mongo_filter: int

    validation.expect(M, CustomUseOfMappingReservedFieldError, mongo_filter=1)


def test_same_id_creation(std_mongo_boot, std_http: Client):
    r: TestHttpResponse = std_http.post(
        "/users",
        200,
        json={
            "name": "Mark Watney"
        }
    )
    r2: TestHttpResponse = std_http.post(
        "/users",
        400,
        json={
            "id": r.json()["value"]["id"],
            "name": "Mark Watney"
        }
    )

    validation.validate(
        BootProxy.ie().api_indication.recover(DuplicateKeyError, r2.json()),
        DuplicateKeyError
    )



from typing import overload

import pytest
from pykit import validation
from pykit.errors import NotFoundError

from orwynn import Module, Service, sql
from orwynn.boot import Boot
from orwynn.bootscript import Bootscript
from orwynn.di.di import Di
from orwynn.sql import SQL
from orwynn.sql.errors import EmptyExecutionQueueError
from orwynn.sql.shd import SHD
from orwynn.sql.testing import S1Item
from orwynn.sql.utils import SQLUtils


class ItemRepo(Service):
    def __init__(self, sql: SQL) -> None:
        super().__init__()

        self._sql: SQL = sql

    def get_one(
        self,
        id: str,
        shd: SHD,
    ) -> S1Item:
        return SQLUtils.get_one(id, S1Item, shd)

    @overload
    def create_one(
        self,
        shd: None = None,
        *,
        name: str,
        price: float,
    ) -> str: ...

    @overload
    def create_one(
        self,
        shd: SHD,
        *,
        name: str,
        price: float,
    ) -> S1Item: ...

    def create_one(
        self,
        shd: SHD | None = None,
        *,
        name: str,
        price: float,
    ) -> S1Item | str:
        with SHD(self._sql, shd) as _shd:
            item: S1Item = S1Item(name=name, price=price)

            _shd.add(item)

            if _shd.is_manageable:
                _shd.commit()
                _shd.refresh(item)
                return item.id
            else:
                return item

    @overload
    def delete_one(
        self,
        id: str,
        shd: SHD,
    ) -> S1Item: ...

    @overload
    def delete_one(
        self,
        id: str,
        shd: None = None,
    ) -> None: ...

    def delete_one(
        self,
        id: str,
        shd: SHD | None = None,
    ) -> S1Item | None:
        with SHD(self._sql, shd) as _shd:
            item: S1Item = self.get_one(id, _shd)

            _shd.delete(item)

            if _shd.is_manageable:
                _shd.commit()
                return None
            else:
                return item


@pytest.mark.asyncio
async def test_execution_normal(
    create_tables_bootscript: Bootscript,
):
    await Boot.create(
        Module(
            Providers=[ItemRepo],
            imports=[sql.module],
        ),
        bootscripts=[
            create_tables_bootscript,
        ],
        apprc={
            "prod": {
                "SQL": {
                    "database_kind": "sqlite",
                    "database_path": ":memory:"
                }
            }
        }
    )

    item_repo: ItemRepo = Di.ie().find("ItemRepo")

    with SHD.new(Di.ie().find("SQL")) as shd:
        created_item_id: str = item_repo.create_one(name="donut", price=1.2)
        created_item: S1Item = item_repo.get_one(created_item_id, shd)

        validation.expect(
            shd.execute_final,
            EmptyExecutionQueueError,
        )

        assert created_item.name == "donut"
        assert created_item.price == 1.2

        item_repo.delete_one(created_item.id)

        validation.expect(
            shd.execute_final,
            EmptyExecutionQueueError,
        )

        validation.expect(
            item_repo.get_one,
            NotFoundError,
            created_item.id,
            shd,
        )


@pytest.mark.asyncio
async def test_execution_upstream(
    create_tables_bootscript: Bootscript,
):
    await Boot.create(
        Module(
            Providers=[ItemRepo],
            imports=[sql.module],
        ),
        bootscripts=[create_tables_bootscript],
        apprc={
            "prod": {
                "SQL": {
                    "database_kind": "sqlite",
                    "database_path": ":memory:"
                }
            }
        }
    )

    item_repo: ItemRepo = Di.ie().find("ItemRepo")

    with SHD.new(Di.ie().find("SQL")) as shd:
        created_item: S1Item = item_repo.create_one(
            shd, name="donut", price=1.2,
        )
        created_item.price = created_item.price + 0.5

        shd.execute_final()
        shd.refresh(created_item)

        assert isinstance(created_item.id, str)
        assert created_item.name == "donut"
        assert created_item.price == 1.7

        deleted_item_id: str = item_repo.delete_one(created_item.id, shd).id

        shd.execute_final()

        validation.expect(
            item_repo.get_one,
            NotFoundError,
            deleted_item_id,
            shd,
        )



"""List of classes an user can subclass from."""
from orwynn.controller.controller import Controller
from orwynn.errorhandler.errorhandler import ErrorHandler
from orwynn.middleware import Middleware
from orwynn.model.model import Model
from orwynn.service.service import Service

# Note that here listed the most basic classes. E.g. Config is not listed
# since it is a derivative from the Model and on the stage of DI
# it will be checked to find these more specific classes.
SUBCLASSABLES: list[type] = [
    Service,
    Controller,
    Middleware,
    Model,
    ErrorHandler
]

ENVIRONS: set[str] = {
    "ORWYNN_MODE",
    "ORWYNN_ROOT_DIR",
    "ORWYNN_RC_PATH",
    "ORWYNN_IS_CATCH_LOGGING_ENABLED_IN_TESTS"
}
"""
All environment variables the framework supports.
"""



from orwynn.bootscript.callable import BootscriptCallable
from orwynn.bootscript.calltime import CallTime
from orwynn.model import Model


class Bootscript(Model):
    func: BootscriptCallable
    call_time: CallTime

    class Config:
        arbitrary_types_allowed = True



from orwynn.apiversion import ApiVersion
from orwynn.app.app import App
from orwynn.controller.controller import Controller
from orwynn.errorhandler.errorhandler import ErrorHandler
from orwynn.middleware import Middleware
from orwynn.module.module import Module
from orwynn.websocket import WebsocketStack
from orwynn.worker.worker import Worker

from .register.controller import ControllerRegister
from .register.middleware import MiddlewareRegister


class Router(Worker):
    """
    Manages how the requests and responses flow through the app.
    """
    def __init__(
        self,
        app: App,
        *,
        modules: list[Module],
        controllers: list[Controller],
        middleware_arr: list[Middleware],
        exception_handlers: list[ErrorHandler],
        global_http_route: str,
        global_websocket_route: str,
        api_version: ApiVersion
    ) -> None:
        super().__init__()
        self.__app: App = app

        self.__modules: list[Module] = modules
        self.__controllers: list[Controller] = controllers
        self.__middleware_arr: list[Middleware] = middleware_arr
        self.__exception_handlers: list[ErrorHandler] = exception_handlers

        self.__websocket_stack: WebsocketStack = WebsocketStack(
            self.__app._fw_websocket_handler
        )

        self.__middleware_register: MiddlewareRegister = MiddlewareRegister(
            app=self.__app,
            middleware_arr=self.__middleware_arr,
            exception_handlers=self.__exception_handlers,
            websocket_stack=self.__websocket_stack
        )
        self.__controller_register: ControllerRegister = ControllerRegister(
            app=self.__app,
            modules=self.__modules,
            controllers=self.__controllers,
            websocket_stack=self.__websocket_stack,
            global_http_route=global_http_route,
            global_websocket_route=global_websocket_route,
            api_version=api_version
        )

        self.__start_registering()

    def __start_registering(self) -> None:
        # Register middleware, it should be done before the controller's
        # adding due to the special websocket middleware registering
        self.__middleware_register.register_all()

        # And finally controllers
        self.__controller_register.register_all()

        # After all actions the websocket stack needs to call register action.
        # It is populated during middleware and controller registering.
        self.__websocket_stack.register_all()




from pykit import validation
from pykit.validation import validate

from orwynn.controller.controller import Controller
from orwynn.di.collecting.acceptordependencies import (
    collect_dependencies_for_acceptor,
)
from orwynn.di.container import DiContainer
from orwynn.http import HttpController, HttpMiddleware
from orwynn.middleware import GlobalMiddlewareSetup, Middleware
from orwynn.model.model import Model
from orwynn.module.module import Module
from orwynn.proxy.boot import BootProxy
from orwynn.url import URLUtils
from orwynn.websocket import WebsocketController, WebsocketMiddleware


class __CoveredRoutes(Model):
    http: list[str]
    websocket: list[str]


__ModuleToCoveredRoutes = tuple[Module, __CoveredRoutes]


def init_other_acceptors(
    container: DiContainer,
    modules: list[Module],
    global_middleware: GlobalMiddlewareSetup | None = None
) -> None:
    """Populates DI container with initialized non-provider acceptors.

    Attributes:
        container:
            DI container.
        modules:
            List of modules to collect acceptors from.
        global_middleware (optional):
            Global middleware setup setting. None is initialized by default.
    """
    if global_middleware is None:
        global_middleware = {}
    __init_modules(container, modules, global_middleware)


def __init_modules(
    container: DiContainer,
    modules: list[Module],
    global_middleware: GlobalMiddlewareSetup
) -> None:
    # Init global middleware first, and then normally module's defined.
    __init_middleware(
        container=container,
        source=global_middleware
    )

    _init_all_error_handlers(container)

    for module in modules:
        covered_routes: __CoveredRoutes = __init_controllers(container, module)
        __init_middleware(
            container=container,
            source=(module, covered_routes)
        )


def _init_all_error_handlers(
    container: DiContainer
) -> None:
    for Eh in BootProxy.ie().ErrorHandlers:
        container.add(
            Eh(**collect_dependencies_for_acceptor(
                Eh,
                container,
                None
            ))
        )


def __init_controllers(
    container: DiContainer,
    module: Module
) -> __CoveredRoutes:
    http_covered_routes: list[str] = []
    websocket_covered_routes: list[str] = []

    for C in module.Controllers:
        validate(C, Controller)

        if module.route is None:
            raise ValueError(
                f"module {module} has not route and shouldn't have added"
                " any controllers or middleware"
            )

        controller: Controller = C(
            **collect_dependencies_for_acceptor(
                C,
                container,
                module
            )
        )
        final_route: str = URLUtils.join_routes(
            module.route, controller.route
        )

        container.add(
            controller
        )

        if issubclass(C, HttpController):
            http_covered_routes.append(final_route)
        elif issubclass(C, WebsocketController):
            # The final route should be duplicated for each controller's
            # event handler, since they are forming the controller's
            # subroute system
            for subroute in C.get_handler_subroutes():
                websocket_covered_routes.append(
                    URLUtils.join_routes(final_route, subroute)
                )
        elif type(C) is Controller:
            raise TypeError(
                f"abstract Controller instance {C} is disallowed"
            )
        else:
            raise TypeError(
                f"unrecognized controller {C}"
            )

    return __CoveredRoutes(
        http=http_covered_routes,
        websocket=websocket_covered_routes
    )


def __init_middleware(
    container: DiContainer,
    source: __ModuleToCoveredRoutes | GlobalMiddlewareSetup
) -> None:
    """
    Inits a middleware.

    Args:
        container:
        source:
            Module and it's covered routes or global middleware setting.
    """
    if isinstance(source, tuple):
        validation.validate_length(source, 2)
        __init_middleware_from_module(
            container=container,
            module=source[0],
            covered_routes=source[1]
        )
    elif isinstance(source, dict):
        __init_global_middleware(
            container=container,
            setup=source
        )
    else:
        raise TypeError(
            f"unrecognized source object {source}"
        )


def __init_global_middleware(
    container: DiContainer,
    setup: GlobalMiddlewareSetup
) -> None:
    for Middleware_, covered_routes_arr in setup.items():
        __register_middleware_in_container(
            container=container,
            Middleware_=Middleware_,
            covered_routes_arr=covered_routes_arr,
            # Omit the availability check
            module=None
        )


def __init_middleware_from_module(
    *,
    container: DiContainer,
    module: Module,
    covered_routes: __CoveredRoutes
) -> None:
    for Mw in module.Middleware:
        validate(Mw, Middleware)

        if module.route is None:
            raise ValueError(
                f"module {module} has not route and shouldn't have added"
                " any controllers or middleware"
            )

        final_covered_routes: list[str]
        if issubclass(Mw, HttpMiddleware):
            final_covered_routes = covered_routes.http
        elif issubclass(Mw, WebsocketMiddleware):
            final_covered_routes = covered_routes.websocket
        elif type(Mw) is Middleware:
            raise TypeError(
                f"abstract Middleware instance {Mw} is disallowed"
            )
        else:
            raise TypeError(
                f"unrecognized middleware {Mw}"
            )

        __register_middleware_in_container(
            container=container,
            Middleware_=Mw,
            covered_routes_arr=final_covered_routes,
            module=module
        )


def __register_middleware_in_container(
    *,
    container: DiContainer,
    Middleware_: type[Middleware],
    covered_routes_arr: list[str],
    module: Module | None
) -> None:
    """
    Adds a middleware to the container.

    Args:
        ...
        module:
            Module to check dependencies availability from. If None, all
            initialized dependencies will be available.
    """
    container.add(
        Middleware_(
            covered_routes=covered_routes_arr,
            **collect_dependencies_for_acceptor(
                Middleware_,
                container,
                module
            )
        )
    )




from orwynn.error.utils import (
    find_detailed_class_for_exception,
)


class Err1(Exception): pass
class Err1_1(Err1): pass
class Err2(Exception): pass


# NOTE: All passed exception lists are shuffled to ensure internal sorting is
#   performing right.


def test_builtin_exception():
    assert find_detailed_class_for_exception(
        ValueError("hello"),
        [Err1, Err2, ValueError, Exception]
    ) is ValueError




from orwynn.service.service import Service


class Database(Service):
    """Represents a connection to database as well as set of actions with this
    database.

    This is an abstract class and defines core methods which should present on
    both SQL and NoSQL databases. Arguments for these core methods in turn vary
    depending on chosen database kind. For example actions for MongoDB will
    require specifying a collection to work with.

    Anyway, these methods for every service subclassing this one, will be
    incapsulated by Mappings. So the end user will work only with these
    mappings only.
    """



from typing import Any, Callable

from fastapi.dependencies.models import Dependant
from fastapi.dependencies.utils import (
    add_non_field_param_to_dependency,
    add_param_to_fields,
    analyze_param,
    get_param_sub_dependant,
    get_typed_signature,
    is_body_param,
)
from fastapi.utils import get_path_param_names


def get_dependant(
    *,
    path: str,
    call: Callable[..., Any],
    name: str | None = None,
    security_scopes: list[str] | None = None,
    use_cache: bool = True,
) -> Dependant:

    # Copy of FastAPI code in order to add some logic (very bad we know).
    # Custom logic marked with (ORWYNN).
    #
    # Changed 2023-12-03 to comply with FastAPI updates.

    path_param_names = get_path_param_names(path)
    endpoint_signature = get_typed_signature(call)
    signature_params = endpoint_signature.parameters
    dependant = Dependant(
        call=call,
        name=name,
        path=path,
        security_scopes=security_scopes,
        use_cache=use_cache,
    )
    for param_name, param in signature_params.items():
        # (ORWYNN) Ignore orwynn's framework internal arguments
        if param_name.startswith("_fw_"):
            continue

        is_path_param = param_name in path_param_names
        type_annotation, depends, param_field = analyze_param(
            param_name=param_name,
            annotation=param.annotation,
            value=param.default,
            is_path_param=is_path_param,
        )
        if depends is not None:
            sub_dependant = get_param_sub_dependant(
                param_name=param_name,
                depends=depends,
                path=path,
                security_scopes=security_scopes,
            )
            dependant.dependencies.append(sub_dependant)
            continue
        if add_non_field_param_to_dependency(
            param_name=param_name,
            type_annotation=type_annotation,
            dependant=dependant,
        ):
            assert (  # noqa: S101
                    param_field is None
                ), \
                "Cannot specify multiple FastAPI annotations for" \
                + f" {param_name!r}"
            continue
        assert param_field is not None  # noqa: S101
        if is_body_param(param_field=param_field, is_path_param=is_path_param):
            dependant.body_params.append(param_field)
        else:
            add_param_to_fields(field=param_field, dependant=dependant)
    return dependant





import pytest
from pytest import fixture

from orwynn.app.app import App
from orwynn.boot import Boot
from orwynn.module import Module
from orwynn.websocket import Websocket, WebsocketController


@fixture
def std_app(std_boot: Boot) -> App:
    return std_boot.app


@pytest.mark.asyncio
async def test_openapi(run_endpoint):
    data: dict = Boot.ie().app.client.get_jsonify("/openapi.json", 200)

    path: dict = data["paths"]["/"]["get"]
    assert path["deprecated"] is True
    assert path["tags"] == ["best-item", "buy-now"]
    assert \
        path["responses"]["201"]["description"] == "Best response description"
    assert \
        path["responses"]["201"]["description"] == "Best response description"


@pytest.mark.asyncio
async def test_get_dependant_patch_against_websocket():
    """
    Should correctly operate for websocket controllers and builtin middleware.
    """

    # FIXME:
    #   Due to strange reasons, this test works fine in orwynn environment,
    #   but gives an error related to fastapi.dependencies.utils.get_dependant
    #   function, which cannot handle _fw_handlers and other framework's
    #   related arguments.

    class WsCtrl(WebsocketController):
        Route = "/"

        async def main(self, websocket: Websocket) -> None:
            await websocket.send_json({"message": "hello"})

    boot: Boot = await Boot.create(
        Module("/", Controllers=[WsCtrl])
    )

    with boot.app.client.websocket("/") as ws:
        ws.receive_json()


@pytest.mark.asyncio
async def test_custom_docs_route():
    boot: Boot = await Boot.create(
        Module("/"),
        apprc={
            "prod": {
                "App": {
                    "docs_route": "/mydocs"
                }
            }
        }
    )

    # TODO: maybe to add checking that this is truly OpenAPI page returned,
    #       but now i'm too lazy for this
    boot.app.client.get("/mydocs", 200)


@pytest.mark.asyncio
async def test_custom_redoc_route():
    boot: Boot = await Boot.create(
        Module("/"),
        apprc={
            "prod": {
                "App": {
                    "redoc_route": "/myredoc"
                }
            }
        }
    )

    # TODO: maybe to add checking that this is truly ReDoc page returned,
    #       but now i'm too lazy for this
    boot.app.client.get("/myredoc", 200)



# A request of any supported protocol
from typing import TYPE_CHECKING, Union

from orwynn.url import URLMethod, URLScheme

if TYPE_CHECKING:
    from orwynn.http import HttpRequest, HttpResponse
    from orwynn.websocket import Websocket


GenericRequest = Union[
    "HttpRequest",
    # The Websocket object is a representation of the request
    "Websocket"
]
GenericResponse = Union[
    "HttpResponse",
    # The Websocket protocol doesn't return any response in direct form from a
    # controller - it sends it over special channel instead.
    None
]


REQUEST_METHOD_BY_PROTOCOL: dict[URLScheme, list[URLMethod]] = {
    URLScheme.HTTP: [
        URLMethod.Get,
        URLMethod.Post,
        URLMethod.Put,
        URLMethod.Delete,
        URLMethod.Patch,
        URLMethod.Options,
    ],
    URLScheme.Websocket: [
        URLMethod.Websocket
    ]
}



from enum import Enum


class CliCommand(Enum):
    """
    Main command that can be given as a first argument for CLI.

    Attributes:
        Test:
            Run application in test mode.
        Dev:
            Run application in dev mode.
        Prod:
            Run application in prod mode.
    """
    Test = "test"
    Dev = "dev"
    Prod = "prod"



from typing import Callable

from .client import Client
from .embeddedclient import EmbeddedTestClient


def get_log_apprc(check_func: Callable) -> dict:
    return {
        "prod": {
            "Log": {
                "handlers": [
                    {
                        "sink": check_func,
                        "serialize": True
                    }
                ]
            },
            "LogHttpMiddleware": {
                "is_request_logged": True,
                "is_reposne_logged": True
            }
        }
    }


class Writer:
    # NOTE: Try adding "await Log.complete()" if log handlers have no time to
    #   process messages before calling Writer.read().
    def __init__(self):
        self._output = ""

    def write(self, message):
        self._output += message

    def read(self):
        return self._output



import pytest
from pykit import validation

from orwynn.boot.boot import Boot
from orwynn.di.container import DiContainer
from orwynn.di.errors import ProviderAvailabilityError
from orwynn.di.init.acceptors import init_other_acceptors
from orwynn.http import Endpoint, HttpController, HttpMiddleware
from orwynn.module import Module
from orwynn.service import Service
from tests.std.assertion import Assertion


@pytest.mark.asyncio
async def test_std(
    std_di_container: DiContainer,
    std_modules: list[Module]
):
    init_other_acceptors(std_di_container, std_modules)

    for A in Assertion.CollectedOtherAcceptors:
        isinstance(std_di_container.find(A.__name__), A)


@pytest.mark.asyncio
async def test_controller_dependency_unavailability():
    """
    Controller shouldn't be allowed to request unavailable modules.
    """
    class UnavailableService(Service):
        pass

    class Ctrl1(HttpController):
        Route = "/"
        Endpoints = [
            Endpoint(method="get")
        ]

        def __init__(self, sv: UnavailableService) -> None:
            super().__init__()

        def get(self) -> dict:
            return {}

    key_module: Module = Module("/key", Controllers=[Ctrl1])
    unavailable_module: Module = Module(Providers=[UnavailableService])

    await validation.expect_async(
        Boot.create(
            root_module=Module("/", imports=[key_module, unavailable_module])
        ),
        ProviderAvailabilityError
    )


@pytest.mark.asyncio
async def test_middleware_dependency_unavailability():
    """
    Middleware shouldn't be allowed to request unavailable modules.
    """
    class UnavailableService(Service):
        pass

    class Mw1(HttpMiddleware):
        def __init__(
            self, covered_routes: list[str], sv: UnavailableService
        ) -> None:
            super().__init__(covered_routes)

    key_module: Module = Module("/key", Middleware=[Mw1])
    unavailable_module: Module = Module(Providers=[UnavailableService])

    await validation.expect_async(
        Boot.create(
            root_module=Module("/", imports=[key_module, unavailable_module])
        ),
        ProviderAvailabilityError
    )


@pytest.mark.asyncio
async def test_unavailability_imported_but_not_exported():
    """
    Check that a provider is not available if it's not exported from an
    imported module
    """
    class UnavailableService(Service):
        pass

    class Ctrl1(HttpController):
        Route = "/"
        Endpoints = [
            Endpoint(method="get")
        ]

        def __init__(self, sv: UnavailableService) -> None:
            super().__init__()

        def get(self) -> dict:
            return {}

    unavailable_module: Module = Module(Providers=[UnavailableService])
    key_module: Module = Module(
        "/key", Controllers=[Ctrl1], imports=[unavailable_module]
    )

    await validation.expect_async(
        Boot.create(
            root_module=Module("/", imports=[key_module, unavailable_module])
        ),
        ProviderAvailabilityError
    )



from orwynn.mongo.document.testing import SimpleDocument


def test_remove(document_1: SimpleDocument, document_2: SimpleDocument):
    assert document_1.remove() == document_1
    assert len(list(SimpleDocument.find_all())) == 1



class MalfunctionError(Exception):
    """Something wrong with the system workflow.

    Rare error signifies problems with framework's source code.
    """


class ExceptionAlreadyHandledError(Exception):
    """
    The exception class is handled more by one error handler.
    """



from orwynn.websocket import (
    Websocket,
    WebsocketMiddleware,
    WebsocketNextCall,
    WebsocketRequestContextId,
)
from orwynn.websocket.log.logger import WebsocketLogger


class LogWebsocketMiddleware(WebsocketMiddleware):
    """Logs information about a websocket request.

    It's recommended to be outermost (at custom level) middleware.
    """
    def __init__(self, covered_routes: list[str]) -> None:
        super().__init__(covered_routes)

        self.__logger: WebsocketLogger = WebsocketLogger()

    async def process(
        self, request: Websocket, call_next: WebsocketNextCall
    ) -> None:
        request_id: str = WebsocketRequestContextId().get()
        await self.__logger.log_request(
            request,
            request_id
        )

        await call_next(request)

        # TODO: Maybe somehow listen for websocket messages sent over the
        #   channel?





# TODO(ryzhovalex): this test strangely initializes the boot and for the
#   second time calls init_providers directly
# 0
# @pytest.mark.asyncio
# async def test_std(
#     std_struct: Module,
#     std_provider_dependencies_map: ProviderDependenciesMap
# ):
#     await Boot.create(std_struct)
#     container: DiContainer = init_providers(
#         std_provider_dependencies_map
#     )

#     for P in Assertion.COLLECTED_PROVIDERS:
#         isinstance(container.find(P.__name__), P)



import re
from typing import Any, Callable

from pykit import validation


class Middleware:
    """Accesses requests before they're processed and responses before
    they're returned.

    Should implement method process(...) accepting request and object of the
    next function to call. This method should always return response in either
    modified or unmodified state.

    Generally shouldn't reimplement method dispatch(...) as it's decision
    entrypoint.

    Attributes:
        covered_routes:
            List of routes covered by this middleware. You can pass ["*"] to
            process all routes.
    """
    def __init__(self, covered_routes: list[str]) -> None:
        validation.validate_each(
            covered_routes, str, expected_sequence_type=list
        )
        self._covered_routes: list[str] = covered_routes
        self.__is_all_routes_allowed: bool = "*" in covered_routes

        if self.__is_all_routes_allowed and len(covered_routes) != 1:
            raise ValueError(
                "if you pass \"*\" to covered routes, don't add any other"
                " values"
            )

    def __should_process(self, route: str) -> bool:
        if self.__is_all_routes_allowed:
            return True

        for allowed_route in self._covered_routes:
            # FIXME: Here in routes other regex-incompatible symbols may
            #   appear, but i don't know which ones at this point. Maybe we
            #   should add more advanced adjustments (replacements) or error
            #   raising.
            #
            # Replace variable brackets and dashes from kebab-style notations.
            #
            # Also, notice that for paths like "/users/10ci/...." allowed route
            # "/users/{id}" will be also compliant, which may not seemed right
            # but due to external checks before route processed to middleware
            # such types of routes won't appear here (i.e. "/users/10ci/foo"
            # won't be passed to the controller "/users/{id}").
            prepared_route: str = re.sub(
                r"\{.+\}", ".+", allowed_route
            ).replace(
                "-", r"\-"
            )

            route_pattern: re.Pattern = re.compile(
                # This way paths will be covered too -
                # e.g. "/file/{}" matches "/file/home/alex/..."
                prepared_route
            )

            if route_pattern.fullmatch(route):
                return True

        return False

    async def dispatch(
        self,
        request: Any,
        call_next: Callable
    ) -> Any:
        if self.__should_process(request.url.path):
            return await self.process(request, call_next)
        else:
            return await call_next(request)

    async def process(
        self,
        request: Any,
        call_next: Any
    ) -> Any:
        return await call_next(request)



from orwynn.di.constants import BUILTIN_ACCEPTORS


def is_acceptor(Class: type) -> bool:
    """Checks if given class is an Acceptor.

    Args:
        Class:
            Any class to check to.

    Returns:
        Flag signifies if given class is an Acceptor.
    """
    return issubclass(Class, tuple(BUILTIN_ACCEPTORS))



from queue import Empty, Queue
from typing import Self

from pykit.errors import LogicError
from pykit.func import FuncSpec
from sqlalchemy import select
from sqlalchemy.orm import Session

from orwynn.sql.errors import (
    EmptyExecutionQueueError,
    ManageableSHDError,
    NonManageableSHDError,
    SqlToUpstreamUnmatchError,
)
from orwynn.sql.sql import SQL


class SHD:
    """
    Special shell for a session.

    If "shd" is None, a new session is created using the given
    Sql service. In this case on context exit, the session will be
    automatically closed. If "shd" is not None, an error will be raised
    on session closing actions (like commit(), rollback() or close()) to
    not break an upstream session management processes.

    All session-significant actions, such as "add", "delete", "commit" and
    "rollback" is not made for an upstream session, but instead, non-closing
    actions like add() and delete(), are added to an execution queue, which
    can be then invoked at main SHD using execute() method. Such method cannot
    be called by SHD which have been accepted an upstream session, but adding
    new calls to such SHD will be propagated to it's upstream session.

    An example for main thread:
    ```python
    def main():
        my_shd = SHD.new(sql)

        donut = donut_service.create_donut(
            my_shd,
            name="best donut"
        )
        bakery = bakery_service.create_bakery(
            my_shd,
            donut_ids=donut_id
        )
        # note that neither donut nor bakery don't have an
        # id yet, since are not commited and refreshed, due
        # to giving an upstream shd to the services

        some_service.delete_item("1234", my_shd)

        my_shd.execute()
        # call list:
        # shd.add(donut)
        # (ignored) shd.commit()
        # (ignored) shd.refresh(donut)
        #
        # shd.add(bakery)
        # (ignored) shd.commit()
        # (ignored) shd.refresh(bakery)
        #
        # shd.delete(item)

        my_shd.commit()
        my_shd.close()
    ```
    """
    def __init__(
        self,
        # sql object can be None but only if upstream_shd is not None
        sql: SQL | None,
        # shd is not set to None by default since it encourages to consider
        # an upstream shd object (e.g. which came as an argument)
        upstream_shd: Self | None,
    ) -> None:
        self.__is_manageable: bool
        self._target_session: Session

        self._execution_queue: "Queue[FuncSpec]" = Queue()
        self.__upstream_shd: SHD | None = upstream_shd

        if self.__upstream_shd is None:
            if sql is None:
                raise SqlToUpstreamUnmatchError
            self._target_session = sql.session
            self.__is_manageable = True
        else:
            self._target_session = self.__upstream_shd.target_session
            self.__is_manageable = False

        # Mirror methods #
        self.scalar = self._target_session.scalar
        self.scalars = self._target_session.scalars
        self.select = select
        self.get = self._target_session.get
        self.query = self._target_session.query
        self.execute_statement = self._target_session.execute
        ##

    def __enter__(self) -> Self:
        return self

    def __exit__(self, type, value, traceback):
        if self.is_manageable:
            self.close()

    @property
    def target_session(self) -> Session:
        return self._target_session

    @property
    def is_manageable(self) -> bool:
        return self.__is_manageable

    @classmethod
    def new(cls, sql: SQL) -> Self:
        """
        Creates a new manageable SHD without an upstream object.

        Args:
            sql:
                Sql object to work with.

        Returns:
            A new SHD created.
        """
        return cls(sql, None)

    @classmethod
    def inherit(cls, upstream: Self) -> Self:
        """
        Creates a new non-manageable SHD with an upstream object.
        """
        return cls(None, upstream)

    def refresh(self, *args) -> None:
        """
        Refresh one or more tables from database.
        """
        if self.is_manageable:
            for arg in args:
                self._target_session.refresh(arg)
        else:
            raise NonManageableSHDError(
                operation_name="refresh",
            )

    def add(self, instance: object, _warn: bool = True) -> None:
        if self.__is_manageable:
            self._target_session.add(instance, _warn)
        else:
            self._add_to_execution_queue(FuncSpec(
                func=self._target_session.add,
                args=(
                    instance,
                    _warn,
                ),
            ))

    def queue(self, funcspec: FuncSpec) -> None:
        """
        Adds a function to execution queue.

        Useful for addding operations like `parent.children.append(child1)` to
        execute it later without getting an empty queue error.
        """
        if self.is_manageable:
            raise ManageableSHDError(
                operation_name="queue",
            )
        self._add_to_execution_queue(funcspec)

    def delete(self, instance: object) -> None:
        if self.__is_manageable:
            self._target_session.delete(instance)
        else:
            self._add_to_execution_queue(FuncSpec(
                func=self._target_session.delete,
                args=(
                    instance,
                ),
            ))

    def commit(self) -> None:
        if self.__is_manageable:
            self._target_session.commit()
        else:
            raise NonManageableSHDError(
                operation_name="commit",
            )

    def rollback(self) -> None:
        if self.__is_manageable:
            self._target_session.rollback()
        else:
            raise NonManageableSHDError(
                operation_name="rollback",
            )

    def close(self) -> None:
        if self.__is_manageable:
            self._target_session.close()
        else:
            raise NonManageableSHDError(
                operation_name="close",
            )

    def execute(self) -> None:
        """
        Executes all writing functions stored in execution queue.

        Can be called only from manageable SHD.
        """
        if not self.__is_manageable:
            raise NonManageableSHDError(
                operation_name="execute",
            )
        elif self._execution_queue.empty():
            # TODO(ryzhovalex): raising this error here breaks isolation OOP
            #   principle because the caller need to known whether the caller
            #   function executed any execution queue additions, need to think
            #   how to better handle this
            # 0
            raise EmptyExecutionQueueError
        else:
            while True:
                try:
                    func_spec: FuncSpec = self._execution_queue.get_nowait()
                    func_spec.call()
                except Empty:
                    return

    def execute_final(self) -> None:
        """
        Executes queue and calls commit on success.
        """
        self.execute()
        self.commit()

    def _add_to_execution_queue(
        self,
        func_spec: FuncSpec,
    ) -> None:
        """
        Adds a func spec to execution queue of this SHD, or of upstream one if
        it exists.
        """
        if self.__is_manageable:
            self._execution_queue.put_nowait(func_spec)
        elif self.__upstream_shd is None:
            raise LogicError(
                f"shd={self} is not manageable, but an upstream shd is None",
            )
        else:
            self.__upstream_shd._add_to_execution_queue(
                func_spec,
            )



from typing import Any

from pykit.env import EnvSpec

from orwynn.config import Config
from orwynn.error import MalfunctionError
from sqlalchemy.pool import StaticPool, Pool

from orwynn.sql.poolclass import PoolclassStr
from .enums import SQLDatabaseKind


class SQLConfig(Config):
    database_kind: SQLDatabaseKind = SQLDatabaseKind.SQLite
    database_name: str | None = None
    database_user: str | None = None
    database_password: str | None = None
    database_path: str = ":memory:"
    database_host: str | None = None
    database_port: int | None = None
    poolclass: PoolclassStr | None = None
    pool_size: int | None = None
    should_drop_env_spec: EnvSpec | None = None

    def __init__(self, **data: Any) -> None:
        try:
            db_kind: SQLDatabaseKind = SQLDatabaseKind(data["database_kind"])
        except KeyError:
            # memory sqlite is used by default
            db_kind = SQLDatabaseKind.SQLite
            data["database_path"] = ":memory:"

        # Check right associations
        if db_kind is SQLDatabaseKind.PostgreSQL:
            for key in [
                "database_name",
                "database_user",
                "database_password",
                "database_host",
                "database_port"
            ]:
                if not data.get(key, None):
                    raise ValueError(
                        f"for PostgreSQL you should define {key} in SQL config"
                    )
        elif db_kind is SQLDatabaseKind.SQLite:
            if not data.get("database_path", None):
                raise ValueError(
                    "for SQLite you should define database_path in SQL config"
                )
        else:
            raise MalfunctionError()

        super().__init__(**data)

    @property
    def real_poolclass(self) -> type[Pool] | None:
        if self.poolclass is None:
            return None

        match self.poolclass:
            case "StaticPool":
                return StaticPool
            case _:
                raise ValueError()



from contextvars import ContextVar, Token
from typing import Any

from pykit import validation

from orwynn.worker.worker import Worker

from .errors import AlreadyInitializedStorageError, UndefinedStorageError


class ContextStorage(Worker):
    def __init__(self) -> None:
        self.__storage: ContextVar[dict | None] = ContextVar(
            "context_storage",
            default=None
        )

    def __get_storage_data_or_error(self) -> dict:
        storage_data: dict | None = self.__storage.get()

        if storage_data is None:
            raise UndefinedStorageError()

        return storage_data

    def init_data(
        self,
        initial_data: dict
    ) -> Token:
        """Initializes the context storage.

        Args:
            initial_data:
                Initial dictionary to set for the storage.

        Returns:
            Context token of the created storage.

        Raises:
            AlreadyInitializedStorageError:
                The storage is already initialized.
        """
        validation.validate(initial_data, dict)

        if self.__storage.get() is not None:
            raise AlreadyInitializedStorageError()

        return self.__storage.set(initial_data)

    def reset(
        self,
        token: Token
    ) -> None:
        """Resets the storage set related to given token.

        Args:
            token:
                Token related to storage's set to be reset.
        """
        validation.validate(token, Token)
        self.__storage.reset(token)

    def get(self, key: Any) -> Any:
        """Retrieves the value from storage by the key.

        Args:
            key:
                Key of the field to be retrieved.

        Returns:
            The value retrieved.

        Raises:
            UndefinedStorageError:
                Storage is not defined for current context.
            KeyError:
                No such key in storage.
        """
        storage_data: dict = self.__get_storage_data_or_error()
        try:
            return storage_data[key]
        except KeyError as err:
            raise KeyError(
                f"no such key {key} in storage"
            ) from err

    def save(self, key: Any, value: Any) -> None:
        """Saves a value for the key into the storage for the current context.

        Args:
            key:
                Key of the field to be saved.
            value:
                Value of the field to be saved.

        Raises:
            UndefinedStorageError:
                Storage is not defined for current context.
        """
        self.__get_storage_data_or_error()[key] = value



from orwynn.indication.indication import Indication
from orwynn.indication.indicator import Indicator
from orwynn.indication.type import IndicationType
from orwynn.model.model import Model


def test_default_indication_type():
    class Item(Model):
        name: str
        price: float

    i: Indication = Indication({
        "type": Indicator.Type,
        "value": Indicator.Value
    })

    assert i.digest(Item(name="pizza", price=2.3))["type"] == "ok"


def test_custom_indication_type():
    class Item(Model):
        _IndicationType = IndicationType.Error
        name: str
        price: float

    i: Indication = Indication({
        "type": Indicator.Type,
        "value": Indicator.Value
    })

    assert i.digest(Item(name="pizza", price=2.3))["type"] == "error"



"""
Scalable web-framework with out-of-the-box architecture.
"""

import importlib.metadata

from orwynn.config import Config
from orwynn.controller import Controller
from orwynn.database import Database
from orwynn.dto import (
    DTO,
    ContainerDTO,
    DTOUtils,
    UnitDTO,
    WebsocketCallDTO,
)
from orwynn.errorhandler import ErrorHandler
from orwynn.middleware import Middleware
from orwynn.model import Model
from orwynn.module import Module
from orwynn.service import Service
from orwynn.worker import Worker

__version__ = importlib.metadata.version("orwynn")

__all__ = [
    "Config",
    "Controller",
    "Database",
    "DTO",
    "ContainerDTO",
    "DTOUtils",
    "UnitDTO",
    "WebsocketCallDTO",
    "ErrorHandler",
    "Middleware",
    "Model",
    "Module",
    "Service",
    "Worker"
]



from contextlib import contextmanager
from contextvars import Token

from pykit import validation

from orwynn.context.storage import ContextStorage


@contextmanager
def context_manager(
    data: dict | None = None
):
    """Populates the context storage with the given data.

    Args:
        data (optional):
            Dict to populate the created storage with. Defaults to empty dict.
    """
    if data is None:
        data = {}

    validation.validate(data, dict)

    storage: ContextStorage = ContextStorage.ie()
    token: Token = storage.init_data(data)

    try:
        yield
    finally:
        storage.reset(token)



import pytest

from orwynn.apiversion import ApiVersion
from orwynn.boot import Boot
from orwynn.http import Endpoint, HttpController
from orwynn.module import Module


@pytest.mark.asyncio
async def test_default():
    """
    By default a client should use the global route.
    """
    class C(HttpController):
        Route = "/message"
        Endpoints = [Endpoint(method="get")]

        def get(self) -> dict:
            return {"message": "hello"}

    boot: Boot = await Boot.create(
        root_module=Module("/user", Controllers=[C]),
        global_http_route="/donuts"
    )

    boot.app.client.get_jsonify("/user/message", 200)


@pytest.mark.asyncio
async def test_default_version():
    """
    By default a client should use the latest api version available.
    """
    class C(HttpController):
        Route = "/message"
        Endpoints = [Endpoint(method="get")]
        Version = 3

        def get(self) -> dict:
            return {"message": "hello"}

    boot: Boot = await Boot.create(
        root_module=Module("/user", Controllers=[C]),
        global_http_route="/api/v{version}",
        api_version=ApiVersion(
            supported={1, 2, 3}
        )
    )

    boot.app.client.get_jsonify("/user/message", 200)


@pytest.mark.asyncio
async def test_not_used():
    """
    The global route can be disabled.
    """
    class C(HttpController):
        Route = "/message"
        Endpoints = [Endpoint(method="get")]

        def get(self) -> dict:
            return {"message": "hello"}

    boot: Boot = await Boot.create(
        root_module=Module("/user", Controllers=[C]),
        global_http_route="/donuts"
    )

    boot.app.client.get_jsonify(
        "/donuts/user/message",
        200,
        is_global_route_used=False
    )


@pytest.mark.asyncio
async def test_pass_version():
    """
    A client is able to not specify global route, but pass own api version.
    """
    class C(HttpController):
        Route = "/message"
        Endpoints = [Endpoint(method="get")]
        Version = 2

        def get(self) -> dict:
            return {"message": "hello"}

    boot: Boot = await Boot.create(
        root_module=Module("/user", Controllers=[C]),
        global_http_route="/api/v{version}",
        api_version=ApiVersion(
            supported={1, 2, 3}
        )
    )

    boot.app.client.get_jsonify(
        "/user/message",
        404,
        api_version=1
    )
    boot.app.client.get_jsonify(
        "/user/message",
        200,
        api_version=2
    )
    boot.app.client.get_jsonify(
        "/user/message",
        404,
        api_version=3
    )



import inspect
from typing import Callable, Coroutine

from orwynn.di.availability import check_availability
from orwynn.di.container import DiContainer
from orwynn.di.object import DiObject
from orwynn.di.provider import Provider
from orwynn.middleware import Middleware
from orwynn.module import Module


def collect_dependencies_for_acceptor(
    acceptor_callable: Callable | Coroutine,
    container: DiContainer,
    acceptor_module: Module | None
) -> dict[str, Provider]:
    """
    Collects all dependencies for given acceptor.

    Note that availabily check won't be performed in any of two cases:
    - acceptor_module is None
    - acceptor_callable is not a class

    Args:
        acceptor_callable:
            Callable acceptor object to inspect.
        container:
            DI container to operate with.
        acceptor_module:
            Module to check dependencies availability from. If None, the
            availability check won't be performed.
    """
    result: dict[str, Provider] = {}
    for param in inspect.signature(acceptor_callable).parameters.values():
        # Skip special case for middleware
        if (
            param.name == "covered_routes"
            and inspect.isclass(acceptor_callable)
            and issubclass(acceptor_callable, Middleware)
        ):
            continue

        dependency: DiObject = container.find(param.annotation.__name__)
        result[param.name] = dependency

        if (
            acceptor_module is not None
            and inspect.isclass(acceptor_callable)
        ):
            check_availability(
                acceptor_callable,
                type(dependency),
                acceptor_module
            )

    return result



class Service:
    """Manages incoming calls to operate on attached business domain.

    Service is a Provider, so it can be injectable. It has higher priority than
    Controllers, since Service never ever need to call a controller, so it
    doesn't have to know about them.
    """



from orwynn.context.manager import context_manager
from orwynn.http.middleware.builtinmiddleware import BuiltinHttpMiddleware
from orwynn.http.middleware.nextcall import HttpNextCall
from orwynn.http.requests import HttpRequest
from orwynn.http.responses import HttpResponse


class ContextBuiltinMiddleware(BuiltinHttpMiddleware):
    """
    Creates a shared context storage active within applied request-response
    cycle.
    """
    async def process(
        self, request: HttpRequest, call_next: HttpNextCall
    ) -> HttpResponse:
        with context_manager():
            return await call_next(request)



import os

import pytest
import pytest_asyncio
from pykit import validation

from orwynn.app import AppMode
from orwynn.apprc.apprc import AppRc
from orwynn.boot.boot import Boot
from orwynn.di.di import Di
from orwynn.http import Endpoint, HttpController
from orwynn.module.errors import CircularDependencyError
from orwynn.module.module import Module
from orwynn.mongo.mongo import Mongo
from orwynn.proxy.boot import BootProxy
from orwynn.service.service import Service
from tests.std.text import TextConfig


class _GService(Service):
    def calculate(self, *args: int) -> int:
        return sum(args)


@pytest_asyncio.fixture
async def std_boot(std_struct: Module) -> Boot:
    return await Boot.create(
        root_module=std_struct
    )


@pytest_asyncio.fixture
async def run_std(std_struct: Module):
    await Boot.create(std_struct)


@pytest_asyncio.fixture
async def std_mongo_boot(std_struct: Module) -> Boot:
    return await Boot.create(
        root_module=std_struct
    )


@pytest.fixture
def set_prod_mode():
    os.environ["ORWYNN_MODE"] = "prod"


@pytest.fixture
def set_dev_mode():
    os.environ["ORWYNN_MODE"] = "dev"


@pytest.fixture
def set_test_mode():
    os.environ["ORWYNN_MODE"] = "test"


@pytest.fixture
def set_std_apprc_path_env() -> None:
    os.environ["ORWYNN_RC_PATH"] = os.path.join(
        os.getcwd(),
        "tests/std/orwynn.yml"
    )


@pytest.mark.asyncio
async def test_init_mode_default(std_struct: Module):
    """
    Default mode should be DEV.
    """
    os.environ["ORWYNN_MODE"] = ""
    boot: Boot = await Boot.create(
        root_module=std_struct,
        apprc={
            "dev": {
                "Mongo": {
                    "url": "mongodb://localhost:9006",
                    "database_name": "orwynn_test"
                }
            }
        }
    )
    assert boot.mode == AppMode.DEV


@pytest.mark.asyncio
async def test_init_mode_test(std_struct: Module):
    os.environ["ORWYNN_MODE"] = "test"
    boot: Boot = await Boot.create(
        root_module=std_struct,
        apprc={
            "test": {
                "Mongo": {
                    "url": "mongodb://localhost:9006",
                    "database_name": "orwynn_test"
                }
            }
        }
    )
    assert boot.mode == AppMode.TEST


@pytest.mark.asyncio
async def test_init_mode_dev(std_struct: Module):
    os.environ["ORWYNN_MODE"] = "dev"
    boot: Boot = await Boot.create(
        root_module=std_struct,
        apprc={
            "dev": {
                "Mongo": {
                    "url": "mongodb://localhost:9006",
                    "database_name": "orwynn_test"
                }
            }
        }
    )
    assert boot.mode == AppMode.DEV


@pytest.mark.asyncio
async def test_init_mode_prod(std_struct: Module):
    os.environ["ORWYNN_MODE"] = "prod"
    boot: Boot = await Boot.create(
        root_module=std_struct,
        apprc={
            "prod": {
                "Mongo": {
                    "url": "mongodb://localhost:9006",
                    "database_name": "orwynn_test"
                }
            }
        }
    )
    assert boot.mode == AppMode.PROD


@pytest.mark.asyncio
async def test_init_incorrect_mode(std_struct: Module):
    os.environ["ORWYNN_MODE"] = "helloworld"
    await validation.expect_async(
        Boot.create(root_module=std_struct),
        ValueError
    )


@pytest.mark.asyncio
async def test_init_enable_mongo(std_struct: Module, set_std_apprc_path_env):
    await Boot.create(
        root_module=std_struct
    )

    validation.validate(Di.ie().find("Mongo"), Mongo)


@pytest.mark.asyncio
async def test_nested_configs_prod(
    std_struct: Module,
    set_std_apprc_path_env,
    set_prod_mode
):
    await Boot.create(
        root_module=std_struct
    )
    app_rc: AppRc = BootProxy.ie().apprc
    text_config: TextConfig = Di.ie().find("TextConfig")

    assert app_rc["Text"]["words_amount"] == text_config.words_amount == 1


@pytest.mark.asyncio
async def test_nested_configs_dev(
    std_struct: Module,
    set_std_apprc_path_env,
    set_dev_mode
):
    await Boot.create(
        root_module=std_struct
    )
    app_rc: AppRc = BootProxy.ie().apprc
    text_config: TextConfig = Di.ie().find("TextConfig")

    assert app_rc["Text"]["words_amount"] == text_config.words_amount == 2


@pytest.mark.asyncio
async def test_nested_configs_test(
    std_struct: Module,
    set_std_apprc_path_env,
    set_test_mode
):
    await Boot.create(
        root_module=std_struct
    )
    app_rc: AppRc = BootProxy.ie().apprc
    text_config: TextConfig = Di.ie().find("TextConfig")

    assert app_rc["Text"]["words_amount"] == text_config.words_amount == 3


@pytest.fixture
def __gmodule() -> Module:
    return Module(
        "/gmodule",
        Providers=[_GService],
        exports=[_GService]
    )

@pytest.mark.asyncio
async def test_global_modules(
    __gmodule: Module
):
    class C1(HttpController):
        Route = "/"
        Endpoints = [Endpoint(method="get")]

        def __init__(self, gservice: _GService) -> None:
            super().__init__()
            self.__gservice: _GService = gservice

        def get(self) -> dict:
            return {"value": self.__gservice.calculate(1, 2, 3)}

    boot: Boot = await Boot.create(
        root_module=Module("/", Controllers=[C1]),
        global_modules=[__gmodule]
    )

    data: dict = boot.app.client.get_jsonify("/", 200)

    assert data["value"] == 6


@pytest.mark.asyncio
async def test_global_modules_reimported(
    __gmodule: Module
):
    """No module can import globally-defined modules."""
    class C1(HttpController):
        Route = "/"
        Endpoints = [Endpoint(method="get")]

        def __init__(self, gservice: _GService) -> None:
            super().__init__()
            self.__gservice: _GService = gservice

        def get(self) -> dict:
            return {"value": self.__gservice.calculate(1, 2, 3)}

    await validation.expect_async(
        Boot.create(
            # Root module reimports globally defined module
            root_module=Module("/", Controllers=[C1], imports=[__gmodule]),
            global_modules=[__gmodule]
        ),
        CircularDependencyError
    )



class DtoError(Exception):
    """
    @abstract
    """


class HasCodeButNoAttributeDtoError(DtoError):
    """
    When a dto does not have `Code` attribute set, but got `code` field.
    """
    def __init__(
        self,
        *,
        code_field: str,
    ):
        message: str = \
            "dto does not have Code attribute but does have `code` field=" \
            f"{code_field}"
        super().__init__(message)


class UnmatchedCodeFieldDtoError(DtoError):
    """
    When dto's `code` field does not match Code class attribute.
    """
    def __init__(
        self,
        *,
        code_field: str,
        code_attribute: str,
    ):
        message: str = \
            f"explicitly given code field={code_field}" \
            f" does not match a class's defined attribute={code_attribute}"
        super().__init__(message)


class AbstractDtoClassAsBaseDtoError(DtoError):
    """
    Container cannot declare base dto class as BASE attribute.
    """
    def __init__(
        self,
    ):
        message: str = "container cannot declare abstract Dto class as BASE"
        super().__init__(message)



def find_detailed_class_for_exception(
    exception: Exception,
    Exceptions: list[type[Exception]]
) -> type[Exception]:
    """
    Searches for the most detailed exception class for which occured exception
    is an instance.

    Args:
        exception:
            Exception to search for.
        Exceptions:
            List of exceptions to search within. It can be previously sorted
            by detail level to optimize if it is possible.

    Returns:
        Parent Exception class with the most detailed level.
    """
    # Get the most detailed (last in the sorted list) class which is also a
    # parent of the given instance.
    return __sort_by_detail([
        Exc for Exc in Exceptions if isinstance(exception, Exc)
    ])[-1]


def __sort_by_detail(
    Exceptions: list[type[Exception]]
) -> list[type[Exception]]:
    """
    Returns a new sorted by detail (lower index -> lower detail) list of
    exceptions.
    """
    copied: list[type[Exception]] = Exceptions.copy()

    length: int = len(Exceptions)
    is_swapped: bool = False
    for i in range(length):
        for j in range(0, length - i - 1):
            if __is_more_detailed(copied[j], copied[j+1]):
                is_swapped = True
                copied[j], copied[j+1] = copied[j+1], copied[j]
        # Exit the main loop after 0 swaps
        if not is_swapped:
            break

    return copied


def __is_more_detailed(
    Exception1: type[Exception],
    Exception2: type[Exception]
) -> bool:
    """
    Returns True if an Exception1 has more detail level than Exception2 and
    False otherwise. Also returns False if both given exception classes have
    the same level of detail.
    """
    mro1: list[type] = Exception1.mro()

    return (Exception2 in mro1)


def get_exception_direct_subclasses() -> list[type[Exception]]:
    """
    Returns all base Exception direct subclasses.
    """
    all_exceptions: list[type[Exception]] = Exception.__subclasses__()
    return all_exceptions



from orwynn.websocket.middleware.middleware import (
    WebsocketMiddleware,
)


class BuiltinWebsocketMiddleware(WebsocketMiddleware):
    """Special framework's middleware which covers all possible routes."""
    def __init__(self) -> None:
        # Cover all routes
        super().__init__(["*"])



from collections.abc import Callable, Sequence
from typing import Any, ClassVar, Self, TypeVar

from pykit import validation
from pykit.cls import ClassUtils, Static
from pykit.errors import NotFoundError, RequiredClassAttributeError
from pykit.types import T

from orwynn.dto.errors import (
    AbstractDtoClassAsBaseDtoError,
    HasCodeButNoAttributeDtoError,
    UnmatchedCodeFieldDtoError,
)
from orwynn.model import Model
from orwynn.websocket import Websocket

_TTable = TypeVar("_TTable")
_TUnitDTO = TypeVar("_TUnitDTO", bound="UnitDTO")
DtoTypes = dict[str, type[_TUnitDTO]]


class DTO(Model):
    """
    Abstract class for Data Transfer Object.
    """


class UnitDTO(DTO):
    """
    Represents some business unit wrapped in DTO to be returned outside.

    Note that model's field "code" is set internally only according to
    class attribute "Code" (see below).

    So on given "code" field explicitly on model's creation, the
    ValueError will be raised, if the given code is not matched class's Code.
    In any other case, this argument will be just ignored.

    Class-attributes:
        Code:
            Type of the dto which will be set at serialization time to field
            "type". Useful to represent different types in interfaces which
            act polymorphically and may return different types of the same
            parent. Defaults to None, i.e. not set.
    """
    Code: ClassVar[str | None] = None
    id: str
    code: str | None = None

    def __init__(self, **data: Any) -> None:
        if "code" in data and data["code"] is not None:
            if self.Code is None:
                raise HasCodeButNoAttributeDtoError(
                    code_field=data["code"],
                )
            elif data["code"] != self.Code:
                raise UnmatchedCodeFieldDtoError(
                    code_field=data["code"],
                    code_attribute=self.Code,
                )

        data["code"] = self.Code

        super().__init__(**data)


class ContainerDTO(DTO):
    """
    Holds a sequence of dto units.

    Class-attributes:
        BASE:
            Instances of which base type (or it's subclasses) will be stored
            in the container. Accepts only UnitDto.
    """
    Base: ClassVar[type[UnitDTO] | None] = None
    units: Sequence[UnitDTO]

    @classmethod
    def _get_base_attr(cls) -> type[UnitDTO]:
        if cls.Base is None:
            raise RequiredClassAttributeError(
                attribute_name="BASE",
                Class=cls,
            )
        elif cls.Base is DTO:
            raise AbstractDtoClassAsBaseDtoError
        else:
            return cls.Base

    @classmethod
    def convert(
        cls,
        objs: list[T],
        convertion_func: Callable[[T], _TUnitDTO],
        container_kwargs: dict[str, Any] | None = None,
    ) -> Self:
        Base: type[UnitDTO] = cls._get_base_attr()

        result: list[_TUnitDTO] = []

        for obj in objs:
            converted: _TUnitDTO = convertion_func(obj)
            validation.validate(converted, Base)
            result.append(converted)

        if not container_kwargs:
            container_kwargs = {}
        return cls(units=result, **container_kwargs)

    @classmethod
    def recover_polymorph(
        cls,
        data: dict,
    ) -> Self:
        """
        Recovers using polymorph strategy setting self as a base parent.

        Args:
            data:
                Dto-compliant raw data.
        """
        Base: type[UnitDTO] = cls._get_base_attr()

        return cls(
            units=DTOUtils.recover_polymorph_units(
                Base=Base,
                units=data["value"]["units"],
            ),
        )


class WebsocketCallDTO(DTO):
    """
    Any call passed in websocket protocol.

    It could be either client's or server's origin.
    """
    websocket: Websocket

    class Config:
        arbitrary_types_allowed = True


class DTOUtils(Static):
    @classmethod
    def recover_polymorph_units(
        cls,
        Base: type[_TUnitDTO],
        units: list[dict],
    ) -> Sequence[_TUnitDTO]:
        """
        Recovers list of item dtos of various types using helper dictionary.

        When some Container DTO cannot be recovered in a simple way using
        Model.recover() because of different UnitDto types which can be stored
        inside, this helper function can be used to retrieve correct list of
        different units for such container.

        This helper function will retrieve some subclass from the given Base,
        where TYPE attribute matches specified one in `units` for each item.

        Args:
            Base:
                Base parent class of all dto types that possibly can be
                recovered.
            raw_units:
                Dictionaries with raw data to be processed back into dtos.

        Returns:
            List of recovered item dtos.

        Raises:
            NotFoundError:
                Cannot find code for item's type.
            NotFoundError:
                Dto subclass for given base and item's type is not found.

        Usage example:
        ```python
        units: Sequence[AnimalDto] = recover_polymorph_units(
            Base=AnimalDto,
            units=[
                {
                    "id": "ed11",
                    "type": "cat",
                    "meows": 10
                },
                {
                    "id": "vj12",
                    "type": "dog",
                    "barks": 15
                }
            ]
        )

        animals_dto: AnimalsDto = AnimalsDto(units=units)

        print(animals_dto.units)
        # =>
        # [CatDto(...), DogDto(...)]
        ```
        """
        validation.validate_each(units, dict, expected_sequence_type=list)

        dtos: Sequence[_TUnitDTO] = []

        for raw_item in units:
            validation.validate(raw_item, dict)

            try:
                type_code: str = raw_item["code"]
            except ValueError as error:
                raise NotFoundError(
                    title="item code",
                    value=raw_item["code"],
                ) from error

            DtoClass: type[_TUnitDTO] = cls.find_dto_class_with_type_code(
                type_code,
                Base=Base,
            )

            dto: _TUnitDTO = validation.apply(
                DtoClass.parse_obj(raw_item),
                Base,
            )

            dtos.append(dto)

        return dtos

    @classmethod
    def find_dto_class_with_type_code(
        cls,
        type_code: str,
        Base: type[_TUnitDTO],
    ) -> type[_TUnitDTO]:
        """
        Searches subclasses of the given Base and returns one found with the
        same type as given one.
        """
        base_subclasses: list[type[_TUnitDTO]] = \
            ClassUtils.find_all_subclasses(Base)  # type: ignore

        for subclass in base_subclasses:
            if (
                subclass.Code is not None
                and subclass.Code == type_code
            ):
                return subclass

        raise NotFoundError(
            title="dto subclass",
            options={
                "base": Base,
                "type_code": type_code,
            },
        )



import contextlib

from orwynn.url import URLVars
from orwynn.websocket.websocket import Websocket

from .handlers import DispatchWebsocketHandler, WebsocketHandler
from .helpers import get_handler_kwargs


class NextCallHandler:
    """
    Propagates calls through the given handler list.
    """
    def __init__(
        self,
        handlers: list[WebsocketHandler],
        url_vars: URLVars
    ) -> None:
        # Index of function being executed
        self.__current_index: int = 0
        self.__handlers: list[WebsocketHandler] = handlers
        self.__url_vars: URLVars = url_vars

    async def __call__(self, websocket: Websocket) -> None:
        """
        Propagates a call to the next recipient in the function list.
        """
        # Supress: End of the stack, just pass
        with contextlib.suppress(KeyError):
            current_handler: WebsocketHandler = \
                self.__handlers[self.__current_index]
            # For dispatch handlers recreate self and delegate execution right
            # to the new instance using rest available handlers
            if isinstance(current_handler, DispatchWebsocketHandler):
                if self.__current_index == len(self.__handlers) - 1:
                    raise ValueError(
                        "last element of handlers cannot be a dispatch handler"
                        f" {current_handler}"
                    )

                await current_handler.func(
                    websocket,
                    self.__class__(
                        handlers=self.__handlers[self.__current_index+1:],
                        url_vars=self.__url_vars
                    )
                )
            else:
                await current_handler.func(
                    websocket,
                    **get_handler_kwargs(current_handler, self.__url_vars)
                )



import pytest
from pykit import validation
from pykit.errors import NotFoundError

from orwynn.di.di import Di
from orwynn.sql.search import TableSearch
from orwynn.sql.shd import SHD
from orwynn.sql.testing import S1Item
from orwynn.sql.utils import SHDUtils


@pytest.mark.asyncio
async def test_s1(s1_table_search: TableSearch):
    with SHD.new(Di.ie().find("SQL")) as shd:
        validation.expect(
            SHDUtils.scalars_all,
            NotFoundError,
            shd.select(S1Item),
            s1_table_search,
        )



from pykit import validation

from orwynn.errorhandler.errorhandler import ErrorHandler
from orwynn.http.middleware.builtinmiddleware import BuiltinHttpMiddleware
from orwynn.http.middleware.nextcall import HttpNextCall
from orwynn.http.requests import HttpRequest
from orwynn.http.responses import HttpResponse


class ErrorHandlerHttpMiddleware(BuiltinHttpMiddleware):
    """
    Handles all errors occured at Http layer.
    """
    def __init__(
        self,
        handlers: set[ErrorHandler]
    ) -> None:
        super().__init__()
        validation.validate_each(
            handlers, ErrorHandler, expected_sequence_type=set
        )
        self.__handlers: set[ErrorHandler] = handlers

    @property
    def handlers(self) -> set[ErrorHandler]:
        return self.__handlers.copy()

    async def process(
        self,
        request: HttpRequest,
        call_next: HttpNextCall
    ) -> HttpResponse:
        # Here no actions is done at the moment since handlers are added as
        # Starlette exception middleware.
        return await call_next(request)



from typing import TYPE_CHECKING, Any, ClassVar, Self, TypeVar, Union

import pydantic
from pykit import validation

from orwynn.proxy.indicationonly import ApiIndicationOnlyProxy

if TYPE_CHECKING:
    from orwynn.indication.type import IndicationType

RecoverType = TypeVar("RecoverType", bound="Model")


class Model(pydantic.BaseModel):
    """Basic way to represent a data in the app.

    Attributes:
        _IndicationType (optional):
            Type to be displayed in final response body. Defaults to OK.
    """
    _IndicationType: ClassVar[Union["IndicationType", None]] = None

    def __init__(self, **data: Any) -> None:
        super().__init__(**data)

    def __str__(self) -> str:
        return f"{self.__class__.__name__}({super().__str__()})"

    @property
    def api(self) -> dict:
        """Generates API-complying object using project's defined API
        indication.
        """
        return ApiIndicationOnlyProxy.ie().api_indication.digest(self)

    @classmethod
    def recover(cls, mp: dict) -> Self:
        """Recovers model of this class using dictionary."""
        return validation.apply(
                ApiIndicationOnlyProxy.ie().api_indication.recover(
                cls, mp
            ),
            Model
        )

    @classmethod
    def create_dynamic(cls, name: str, /, **kwargs) -> type[Self]:
        return pydantic.create_model(name, __base__=cls, **kwargs)



import re
from typing import TypeVar

from pykit.validation import validate

from orwynn.config import Config
from orwynn.controller.controller import Controller
from orwynn.di.errors import (
    DiObjectAlreadyInitializedInContainerError,
    FinalizedDiContainerError,
    MissingDiObjectError,
)
from orwynn.di.isprovider import is_provider
from orwynn.di.object import DiObject
from orwynn.error.errors import MalfunctionError
from orwynn.errorhandler.errorhandler import ErrorHandler
from orwynn.helpers.constants import SUBCLASSABLES
from orwynn.middleware import Middleware
from orwynn.model.model import Model

_InnerObj = TypeVar("_InnerObj")


class DiContainer:
    """Holds data about initialized di objects."""
    def __init__(self) -> None:
        self._data: dict[str, DiObject] = {}

        # Struct running in parallel to main data one to hold references by
        # class for self properties like getting all controllers
        self._objects_by_base_class: dict[type[DiObject], list[DiObject]] = {}

        self._is_provider_populating_finalized: bool = False

    @property
    def items(self) -> list[tuple[str, DiObject]]:
        result: list[tuple[str, DiObject]] = []

        for name, obj in self._data.items():
            result.append((name, obj))

        return result

    @property
    def controllers(self) -> list[Controller]:
        """Fetches all controllers from container.

        Returns:
            All controllers fetched.
        """
        result: list[Controller] = self._find_objects_for_class(Controller)
        return result

    @property
    def all_middleware(self) -> list[Middleware]:
        """Fetches all middleware from container.

        Returns:
            All middleware fetched.
        """
        result: list[Middleware] = self._find_objects_for_class(Middleware)
        return result

    @property
    def exception_handlers(self) -> list[ErrorHandler]:
        result: list[ErrorHandler] = self._find_objects_for_class(
            ErrorHandler
        )
        return result

    def add(self, obj: DiObject) -> None:
        """Adds DI object to the container.

        Args:
            obj:
                DI object to add.

        Raises:
            DIObjectAlreadyInitializedError:
                This DI object already exists in container.
        """
        if self._is_provider_populating_finalized and is_provider(type(obj)):
            raise FinalizedDiContainerError(
                f"cannot add provider {obj}, container is finalized"
            )

        obj_class_name: str = obj.__class__.__name__

        if obj_class_name in self._data:
            raise DiObjectAlreadyInitializedInContainerError(
                f"DI object {obj} already initialized"
            )

        self._data[obj_class_name] = obj

        self._assign_obj_to_base_class(obj)

    def find(self, key: str) -> DiObject:
        """Returns DI object by its key.

        Note that searching is made using PascalCased keys, but actual object
        returned is an initialized instance of searched class.

        Args:
            key:
                String value to search with.

        Returns:
            A DIObject found.

        Raises:
            MissingDIObjectError:
                DIObject with given key is not found.
        """
        validate(key, str)

        try:
            return self._data[key]
        except KeyError as error:
            raise MissingDiObjectError(
                f"di object for key \"{key}\" is not found"
            ) from error

    def find_re(self, pattern: str) -> list[DiObject]:
        """Searches di objects by pattern.

        Args:
            pattern:
                Pattern to apply.

        Returns:
            All di objects with keys matched given pattern.
        """
        validate(pattern, str)

        result: list[DiObject] = []

        for k, v in self._data.items():
            if re.match(pattern, k):
                result.append(v)

        if result == []:
            raise MissingDiObjectError(
                "di objects for pattern \"pattern\" are not found"
            )

        return result

    def finalize_provider_populating(
        self
    ) -> None:
        """Finalizes the container applying logic after provider populating.
        """
        if self._is_provider_populating_finalized:
            raise ValueError("already finalized")

        self._is_provider_populating_finalized = True

    def _assign_obj_to_base_class(self, obj: DiObject) -> None:
        ObjCls: type[DiObject] = type(obj)
        BaseCls: type[DiObject] | None = None

        # Find out object's base class
        for C in SUBCLASSABLES:
            if issubclass(ObjCls, C):
                # Choose more specific base classes for subclasses of Model
                if C is Model and issubclass(ObjCls, Config):
                    BaseCls = Config
                else:
                    BaseCls =  C
        if not BaseCls:
            raise MalfunctionError(
                f"cannot found base class for obj {obj}, this object"
                " shouldn't be appeared in container"
            )

        try:
            self._objects_by_base_class[BaseCls].append(obj)
        except KeyError:
            self._objects_by_base_class[BaseCls] = [obj]

    def _find_objects_for_class(self, C: type[_InnerObj]) -> list[_InnerObj]:
        result: list[_InnerObj] = []
        is_missing: bool = False

        try:
            result = self._objects_by_base_class[C]
        except KeyError:
            is_missing = True
        else:
            if result == []:
                is_missing = True

        if is_missing:
            raise MissingDiObjectError(
                f"cannot find any object for class {C}"
            )
        else:
            return result



from orwynn.http.context.id import HttpRequestContextId

from .constants import BUILTIN_HTTP_MIDDLEWARE, DEFAULT_HTTP_ERROR_HANDLERS
from .controller.controller import HttpController
from .controller.endpoint.container import EndpointContainer
from .controller.endpoint.endpoint import Endpoint
from .controller.endpoint.response import EndpointResponse
from .cors.cors import Cors
from .errorhandler.default import (
    DefaultErrorHandler,
    DefaultHttpErrorHandler,
    DefaultRequestValidationErrorHandler,
)
from .errorhandler.middleware import (
    ErrorHandlerHttpMiddleware,
)
from .log.configs import LogHttpMiddlewareConfig
from .log.middleware import LogMiddleware
from .middleware.builtinmiddleware import BuiltinHttpMiddleware
from .middleware.middleware import HttpMiddleware
from .middleware.nextcall import HttpNextCall
from .requests import HttpRequest
from .responses import (
    FileHttpResponse,
    HtmlHttpResponse,
    HttpResponse,
    JsonHttpResponse,
    RedirectHttpResponse,
    TestHttpResponse,
)
from .schema.validationvalue import (
    RequestValidationExceptionValueSchema,
)
from .schema.value import HttpExceptionValueSchema



from datetime import time, timedelta
from typing import Callable

from pykit.types import CashOperator

from orwynn.log.types import LogMessage
from orwynn.model.model import Model


class LogHandler(Model):
    # Only mattering for default values fields are added from loguru, others
    # are moved to kwargs dict. Fields set to None by default will be assigned
    # at runtime depending on some conditions
    sink: str | CashOperator | Callable[[LogMessage], None]
    level: int | str | None = None
    format: str | Callable = \
        "{time:%Y.%m.%d at %H:%M:%S.%f%z}" \
        + " | {level} | {extra} >> {message}"
    # Callable here used instead of loguru.RotationFunction since it has
    # problems with importing
    rotation: str | int | time | timedelta | Callable = "10 MB"
    serialize: bool | None = None
    kwargs: dict | None = None



import json

import pytest

from orwynn.boot.boot import Boot
from orwynn.http import Endpoint, HttpController
from orwynn.log import Log
from orwynn.log.types import LogMessage
from orwynn.module.module import Module
from orwynn.testing import Writer, get_log_apprc
from orwynn.websocket import Websocket, WebsocketController


@pytest.fixture
def writer() -> Writer:
    return Writer()


@pytest.fixture
def log_apprc_sink_to_writer(writer: Writer) -> dict:
    def __check(message: LogMessage) -> None:
        writer.write(message)
    return get_log_apprc(__check)


@pytest.mark.asyncio
async def test_logged_request_id(
    writer: Writer,
    log_apprc_sink_to_writer: dict
):
    class C1(HttpController):
        Route = "/"
        Endpoints = [
            Endpoint(method="get")
        ]

        def get(self) -> dict:
            Log.info("hello")
            return {}

    boot: Boot = await Boot.create(
        Module("/", Controllers=[C1]),
        apprc=log_apprc_sink_to_writer
    )

    boot.app.client.get("/", 200)
    data: dict = json.loads(str(writer.read()))
    extra: dict = data["record"]["extra"]

    assert isinstance(extra["http.request_id"], str)


@pytest.mark.asyncio
async def test_logged_websocket_request_id(
    writer: Writer,
    log_apprc_sink_to_writer: dict
):
    class C1(WebsocketController):
        Route = "/"

        async def main(self, websocket: Websocket) -> None:
            Log.info("hello")

    boot: Boot = await Boot.create(
        Module("/", Controllers=[C1]),
        apprc=log_apprc_sink_to_writer
    )  # type: ignore #worker

    with boot.app.client.websocket("/"):
        pass

    raw: str = str(writer.read())
    assert raw != ""

    data: dict = json.loads(raw)
    extra: dict = data["record"]["extra"]
    assert isinstance(extra["websocket.request_id"], str)



from enum import Enum


class CallTime(Enum):
    """
    When a bootscript will be called.

    Options:
        AFTER_ALL:
            Bootscript will be launched after all framework's boot operations.
    """
    AFTER_ALL = "after_all"



import json

import pytest

from orwynn.boot.boot import Boot
from orwynn.log import Log
from orwynn.module.module import Module
from orwynn.testing import Writer
from orwynn.testing.client import Client
from orwynn.websocket import Websocket, WebsocketController
from orwynn.websocket.log.middleware import LogWebsocketMiddleware


@pytest.mark.asyncio
async def test_get(
    writer: Writer,
    log_apprc_sink_to_writer: dict
):
    class C1(WebsocketController):
        Route = "/"

        async def main(self, websocket: Websocket) -> None:
            await websocket.close()

    boot: Boot = await Boot.create(
        Module(
            route="/", Controllers=[C1], Middleware=[LogWebsocketMiddleware]
        ),
        apprc=log_apprc_sink_to_writer
    )
    client: Client = boot.app.client

    with client.websocket("/"):
        pass

    __check_log_message(writer.read())
    Log.remove()


def __check_log_message(message: str) -> list[dict]:
    assert message != ""
    items: list[str] = message.split("\n")
    parsed_items: list[dict] = []

    for item in items:
        if item == "":
            continue
        data: dict = json.loads(str(item))
        extra: dict = data["record"]["extra"]

        request_data: dict = extra["websocket.request"]
        assert type(request_data["id"]) is str
        assert type(request_data["url"]) is str
        assert type(request_data["headers"]) is dict

        parsed_items.append(data)

    return parsed_items



import contextlib
import os
from copy import deepcopy
from pathlib import Path
from types import NoneType
from typing import Self

import dotenv
from pykit import validation
from pykit.errors import DirectoryExpectError
from pykit.validation import validate, validate_dict, validate_each
from starlette.types import Receive, Scope, Send

from orwynn.apiversion import ApiVersion
from orwynn.app import App, AppMode
from orwynn.apprc import AppRc, AppRCUtils
from orwynn.bootscript import Bootscript, BootscriptWorker, CallTime
from orwynn.bootscript.errors import NoScriptsForCallTimeError
from orwynn.controller import Controller
from orwynn.di.di import Di
from orwynn.di.errors import MissingDiObjectError
from orwynn.errorhandler import ErrorHandler
from orwynn.helpers.errors import DeprecatedFeatureError
from orwynn.http import EndpointContainer
from orwynn.indication import Indication, default_api_indication
from orwynn.log import LogConfig, LogUtils
from orwynn.middleware import GlobalMiddlewareSetup, Middleware
from orwynn.module import Module
from orwynn.proxy.boot import BootProxy
from orwynn.proxy.indicationonly import ApiIndicationOnlyProxy
from orwynn.router import Router
from orwynn.testing import Client
from orwynn.worker import Worker


class Boot(Worker):
    """Worker responsible of booting an application.

    General usage is to construct this class in the main.py with required
    parameters and then access Boot.app for your needs.

    Environs:
        - ORWYNN_MODE:
            Boot mode for application. Defaults to DEV. Alternatively you can
            pass arg "mode".
        - ORWYNN_ROOT_DIR:
            Root directory for application. Defaults to os.getcwd()
        - ORWYNN_RC_PATH:
            Path where app configuration file located. Defaults to
            "./apprc.yml". Alternatively you can pass a dictionary directly in
            "apprc" attribute.
    """
    def __init__(
        self,
        root_module: Module,
        *,
        dotenv_path: Path | None = None,
        api_indication: Indication | None = None,
        ErrorHandlers: set[type[ErrorHandler]] | None = None,
        apprc: AppRc | None = None,
        mode: AppMode | None = None,
        global_http_route: str | None = None,
        global_websocket_route: str | None = None,
        global_modules: list[Module] | None = None,
        global_middleware: GlobalMiddlewareSetup | None = None,
        api_version: ApiVersion | None = None,
        bootscripts: list[Bootscript] | None = None,
        _fw_init_lock: bool = True
    ) -> None:
        self._check_init_lock(_fw_init_lock)

        if dotenv_path is None:
            dotenv_path = Path(".env")
        validate(dotenv_path, Path)
        validate(root_module, Module)
        if api_indication is None:
            api_indication = default_api_indication
        validate(api_indication, Indication)
        if ErrorHandlers is None:
            ErrorHandlers = set()
        validate_each(
            ErrorHandlers, ErrorHandler, expected_sequence_type=set
        )
        validate(apprc, [AppRc, NoneType])
        validate(mode, [AppMode, NoneType])

        if global_http_route is None:
            global_http_route = ""
        validate(global_http_route, str)

        if global_websocket_route is None:
            global_websocket_route = ""
        validate(global_websocket_route, str)

        if global_modules is None:
            global_modules = []
        validate_each(global_modules, Module, expected_sequence_type=list)

        if global_middleware is None:
            global_middleware = {}
        validate_dict(
            global_middleware, (Middleware, list)
        )

        if api_version is None:
            api_version = ApiVersion()
        validate(api_version, ApiVersion)

        if bootscripts is None:
            bootscripts = []
        validate_each(
            bootscripts,
            Bootscript,
            expected_sequence_type=list
        )

        dotenv.load_dotenv(dotenv_path, override=True)

        self.__mode: AppMode = self.__initialize_mode(mode)
        self.__root_dir: Path = self.__parse_root_dir()
        self.__api_indication: Indication = api_indication
        self.__apprc: AppRc = AppRCUtils.parse(
            self.__root_dir,
            self.__mode,
            deepcopy(apprc)
        )
        self.__global_http_route: str = global_http_route
        self.__global_websocket_route: str = global_websocket_route
        self.__api_version: ApiVersion = api_version

        # Initialize bootscript
        self._bootscript_worker: BootscriptWorker = BootscriptWorker(
            bootscripts=bootscripts
        )

        self.__initialize_proxies(
            ErrorHandlers=ErrorHandlers,
            api_indication=api_indication
        )

        self.__di: Di = Di(
            root_module,
            global_modules=global_modules,
            global_middleware=global_middleware
        )

        self.__init_router()

        # Configure logging
        LogUtils.configure_log(
            validation.apply(self.__di.find("LogConfig"), LogConfig),
            app_mode_prod=AppMode.PROD
        )

    async def __call__(
        self, scope: Scope, receive: Receive, send: Send
    ) -> None:
        """
        Propagates a call to the created application.
        """
        await self.app(scope, receive, send)

    @classmethod
    async def create(
        cls,
        root_module: Module,
        *,
        dotenv_path: Path | None = None,
        api_indication: Indication | None = None,
        ErrorHandlers: set[type[ErrorHandler]] | None = None,
        apprc: AppRc | None = None,
        mode: AppMode | None = None,
        global_http_route: str | None = None,
        global_websocket_route: str | None = None,
        global_modules: list[Module] | None = None,
        global_middleware: GlobalMiddlewareSetup | None = None,
        api_version: ApiVersion | None = None,
        bootscripts: list[Bootscript] | None = None
    ) -> Self:
        """
        Creates a Boot instance.

        This should be the only way to initialize a new Boot.

        Idea of using this create() method instead of traditional __init__ is
        to await important asynchronous coroutines every time a fresh boot
        is created.

        Args:
            root_module:
                Root module of the app.
            dotenv_path (optional):
                Path to .env file. Defaults to ".env".
            api_indication (optional):
                Indication object used as a convention for outcoming API
                structures. Defaults to predefined by framework's indication
                convention.
            ErrorHandlers (optional):
                List of exception handlers to add. By default framework adds
                the builtin Exception and orwynn.Error handlers.
            apprc (optional):
                Application configuration. By default environ ORWYNN_RC_PATH
                is checked if this arg is not given.
            mode (optional):
                Application mode. By default environ ORWYNN_MODE is
                checked if this arg is not given.
            global_http_route (optional):
                Global route to be prepended to every controller's route.
                Defaults to no route. Can accept formatting "{version}" for an
                API version to be injected into route.
            global_websocket_route (optional):
                Global route to be prepended to every controller's route.
                Defaults to no route. Can accept formatting "{version}" for an
                API version to be injected into route.
            global_imports (optional):
                Modules available across all other modules imported into the
                application. Note that no other instance can import a globally
                available module.
            global_middleware (optional):
                Map of middleware and its covered routes to apply globally.
                Note that every initialized Provider can be accessed from such
                middleware.
            api_version (optional):
                Object describes API versioning rules for the project. By
                default only v1 is supported.
            bootscripts (optional):
                List of bootscripts to be launched at different points of boot
                time.
        """

        boot: Boot = Boot(
            root_module,
            dotenv_path=dotenv_path,
            api_indication=api_indication,
            ErrorHandlers=ErrorHandlers,
            apprc=apprc,
            mode=mode,
            global_http_route=global_http_route,
            global_websocket_route=global_websocket_route,
            global_modules=global_modules,
            global_middleware=global_middleware,
            api_version=api_version,
            bootscripts=bootscripts,
            _fw_init_lock=False
        )

        await boot._call_bootscripts()

        return boot

    async def _call_bootscripts(self) -> None:
        """
        Calls bootscripts for all time frames.
        """
        # AFTER_ALL bootscript call time
        with contextlib.suppress(NoScriptsForCallTimeError):
            await self._bootscript_worker.call_by_time(
                CallTime.AFTER_ALL,
                self.__di._fw_container
            )

    def __initialize_mode(
        self,
        input_mode: AppMode | None
    ) -> AppMode:
        if input_mode:
            return input_mode
        else:
            return self.__parse_mode()

    def __initialize_proxies(
        self,
        ErrorHandlers: set[type[ErrorHandler]],
        api_indication: Indication
    ) -> None:
        BootProxy(
            root_dir=self.__root_dir,
            mode=self.__mode,
            api_indication=self.__api_indication,
            apprc=self.__apprc,
            ErrorHandlers=ErrorHandlers,
            global_http_route=self.__global_http_route,
            global_websocket_route=self.__global_websocket_route,
            api_version=self.__api_version
        )
        EndpointContainer()
        ApiIndicationOnlyProxy(api_indication)

    def __init_router(self) -> Router:
        try:
            all_modules: list[Module] = self.__di.modules
        except MissingDiObjectError:
            all_modules = []

        try:
            all_controllers: list[Controller] = self.__di.controllers
        except MissingDiObjectError:
            all_controllers = []

        try:
            all_middleware: list[Middleware] = self.__di.all_middleware
        except MissingDiObjectError:
            all_middleware = []

        try:
            all_exception_handlers: list[ErrorHandler] = \
                self.__di.exception_handlers
        except MissingDiObjectError:
            all_exception_handlers = []

        return Router(
            self.app,
            global_http_route=self.__global_http_route,
            global_websocket_route=self.__global_websocket_route,
            api_version=self.__api_version,
            modules=all_modules,
            controllers=all_controllers,
            middleware_arr=all_middleware,
            exception_handlers=all_exception_handlers
        )

    @property
    def app(self) -> App:
        return self.__di.app_service

    @property
    def client(self) -> Client:
        """
        Convenience shortcut to get testing client from the App.
        """
        return self.app.client

    @property
    def mode(self) -> AppMode:
        return self.__mode

    @property
    def api_indication(self) -> Indication:
        return self.__api_indication

    def __parse_mode(self) -> AppMode:
        mode_env: str | None = os.getenv("ORWYNN_MODE")

        if not mode_env:
            return AppMode.DEV
        else:
            return AppMode(mode_env)

    def __parse_root_dir(self) -> Path:
        root_dir: Path
        root_dir_env: str = os.getenv("ORWYNN_ROOT_DIR", "")

        root_dir = \
            Path(os.getcwd()) if not root_dir_env else Path(root_dir_env)

        if not root_dir.is_dir():
            raise DirectoryExpectError(
                path=root_dir
            )

        return root_dir

    def _check_init_lock(self, lock: bool) -> None:
        if lock:
            raise DeprecatedFeatureError(
                deprecated_feature="initializing Boot via __init__",
                use_instead="Boot.create(...)"
            )



import inspect

from pykit.fmt import FormatUtils

from orwynn.config import Config
from orwynn.di.collecting.errors import (
    NoDependenciesForGivenProviderError,
)
from orwynn.di.collecting.providerdependencies.map import (
    ProviderDependenciesMap,
)
from orwynn.di.container import DiContainer
from orwynn.di.errors import MissingDiObjectError
from orwynn.di.isprovider import is_provider
from orwynn.di.provider import Provider
from orwynn.module.errors import CircularDependencyError


def init_providers(
    provider_dependencies_map: ProviderDependenciesMap
) -> DiContainer:
    """Traverses through given providers and dependencies initializing them.

    Args:
        providers_dependencies_map:
            Map of providers and their dependencies.

    Returns:
        DI container populated with initialized providers by their names.
    """
    # Note for this module that most validation logic has been already
    # performed at previous stages, so no need to do checks here.

    container: DiContainer = DiContainer()

    for P in provider_dependencies_map.Providers:
        try:
            container.find(P.__name__)
        except MissingDiObjectError:
            _traverse(
                StarterProvider=P,
                mp=provider_dependencies_map,
                container=container,
                chain=[]
            )

    container.finalize_provider_populating()

    return container


def _traverse(
    *,
    StarterProvider: type[Provider],
    mp: ProviderDependenciesMap,
    container: DiContainer,
    chain: list[type[Provider]]
) -> Provider:
    already_initialized_dependencies: list[Provider] = []

    if StarterProvider in chain:
        raise CircularDependencyError(
            # Failed provider is added second time to the chain for
            # error descriptiveness
            f"provider {StarterProvider} occured twice in dependency chain"
            f" {FormatUtils.format_chain(chain + [StarterProvider])}"
        )

    try:
        Dependencies: list[type[Provider]] = mp.get_dependencies(
            StarterProvider
        )
    except NoDependenciesForGivenProviderError:
        # Final point in chain, since no dependencies we can initialize
        # without any attributes
        pass
    else:
        chain.append(StarterProvider)

        for D in Dependencies:
            try:
                already_initialized_dependencies.append(
                    container.find(D.__name__)
                )
            except MissingDiObjectError:
                already_initialized_dependencies.append(
                    _traverse(
                        StarterProvider=D,
                        mp=mp,
                        container=container,
                        chain=chain
                    )
                )

        # On object initialization it should be removed from chain, but not for
        # Config since it hasn't been appended
        chain.pop()

    # All dependencies for this provider have been initialized - now initialize
    # this provider. Checks on previous phases should have been validated, that
    # is every provider waits only positional arguments.
    result_provider: Provider
    if issubclass(StarterProvider, Config):
        # For configs we need to prepare names kwargs, so we perform
        # signaturing and container searching here once again. It's not good
        # since we've done it before, so consider refactoring.
        provider_kwargs: dict[str, Provider] = {}

        for param in inspect.signature(StarterProvider).parameters.values():
            if (
                inspect.isclass(param.annotation)
                and is_provider(param.annotation)
            ):
                provider_kwargs[param.name] = container.find(
                    param.annotation.__name__
                )

        result_provider = StarterProvider.load(extra=provider_kwargs)
    else:
        result_provider = StarterProvider(
            *already_initialized_dependencies
        )  # type: ignore

    container.add(result_provider)
    return result_provider



from typing import Any
import uvicorn
from orwynn.boot.boot import Boot
from orwynn.helpers.errors import UnsupportedError
from orwynn.server.engine import ServerEngine


class Server:
    """
    Serves the application.

    For now the sole option available - serving via Uvicorn. In future the
    option list might be extended.

    Args:
        engine:
            Underlying server to use.
        boot:
            Orwynn Boot instance.
    """
    def __init__(
        self,
        *,
        engine: ServerEngine,
        boot: Boot
    ) -> None:
        self._engine: ServerEngine = engine
        self._boot: Boot = boot

    async def serve(
        self,
        *,
        host: str,
        port: int
    ) -> None:
        """
        Serves the application from given at initialization Boot instance and
        selected underlying engine.

        Args:
            host:
                Host to serve on.
            port:
                Port to serve on.
        """
        await self._get_core_server(
            host=host,
            port=port
        ).serve()

    def _get_core_server(
        self,
        *,
        host: str,
        port: int
    ) -> Any:
        match self._engine:
            case ServerEngine.Uvicorn:
                return self._get_uvicorn_server(
                    host=host,
                    port=port
                )
            case _:
                raise UnsupportedError(
                    title="server engine",
                    value=self._engine
                )

    def _get_uvicorn_server(
        self,
        *,
        host: str,
        port: int
    ) -> uvicorn.Server:
        config: uvicorn.Config = uvicorn.Config(
            self._boot,
            host=host,
            port=port
        )

        return uvicorn.Server(config)



from __future__ import annotations

import contextlib
from typing import Any, TypeVar

SingletonInstance = TypeVar("SingletonInstance")


class SingletonMeta(type):
    """Singleton metaclass for implementing singleton patterns.

    See:
        https://stackoverflow.com/questions/6760685/creating-a-singleton-in-python
    """
    __instances: dict[type, object] = {}

    def __call__(cls, *args, **kwargs) -> Any:
        if cls not in cls.__instances:
            cls.__instances[cls] = super().__call__(*args, **kwargs)
        return cls.__instances[cls]

    def __validate_in_instances(cls, cannot_message: str) -> None:
        if cls not in cls.__instances.keys():
            raise ValueError(
                f"{cannot_message} - class {cls} not initialized"
            )

    def discard(cls, should_validate: bool = True) -> None:
        if should_validate:
            cls.__validate_in_instances("cannot discard")
        with contextlib.suppress(KeyError):
            del cls.__instances[cls]


class Singleton(metaclass=SingletonMeta):
    """Singleton base class."""
    @classmethod
    def ie(cls: type[SingletonInstance]) -> SingletonInstance:
        """Gets the single instance of the Singleton.

        Returns:
            Instance the Singleton holds.
        """
        return cls()
